unit CommonUnit;

interface

uses
  System.SysUtils, System.Types, System.UITypes, System.Classes, System.Variants,
  FMX.Types, FMX.Graphics, System.DateUtils, System.StrUtils, FMX.Utils,
  System.Net.HttpClient, System.Net.HttpClientComponent, System.Net.URLClient,
  System.NetConsts,
  System.Rtti, System.Math.Vectors, System.Math, FMX.Surfaces, FMX.Forms,
  FMX.Platform, FMX.Clipboard, System.UIConsts, System.IOUtils, System.SyncObjs,
  System.Generics.Collections, System.Generics.Defaults, FMX.TextLayout;

type
  TLogType = (ltError, ltInfo, ltEEW, ltNetError, ltAlert, ltNTP, ltDebug);

  PLogRec = ^TLogRec;
  TLogRec = record
    DateTime: TDateTime;
    LogType: TLogType;
    Message1, Message2: string;
  end;

  TKyoshinIndex =
      (KIacmap_s, KIjma_s, KIvcmap_s, KIdcmap_s,
       KIrsp0125_s, KIrsp0250_s, KIrsp0500_s, KIrsp1000_s,
       KIacmap_b, KIjma_b, KIvcmap_b, KIdcmap_b,
       KIrsp0125_b, KIrsp0250_b,KIrsp0500_b, KIrsp1000_b);

  PEEWRec = ^TEEWRec;
  TEEWRec = record
    KyoshinTime: TDateTime;
    ReportTime: TDateTime;
    ReportID: string;
    ReportNo: integer;
    IsCancel: boolean;
    IsAlert: boolean;
    IsFinal: boolean;
    IsTraining: boolean;
    OriginTIme: TDateTime;
    EpicenterCode: integer;
    EpicenterName: string;
    EpicenterF: TPointF;
    Epicenter: TPoint;
    Depth: string;
    DepthInt: integer;
    Magnitude: single;
    MaxShindo: string;
  end;

  TAverage = record
    DateTime: TDateTime;
    Count: integer;
    Sum: integer;
  end;

  TAverages = record
  strict private
    Average: array[0..3] of TAverage;
    CurrentIndex: integer;
    CurrentTime: TDateTime;
    procedure DoShift(DateTime: TDateTime);
  public
    constructor Create(DateTime: TDateTime);
    function CalcAverage: integer;
    procedure Add(Value: integer; DateTime: TDateTime);
  end;

  PKyoshinRec = ^TKyoshinRec;
  TKyoshinRec = record
    Index: integer;
    Name: string;
    AreaIndex: integer;
    GeoPoint: TPointF;
    ScanPoint: TPoint;
    DispPoint: TPoint;
    Color: TAlphaColor;
    Scale: integer;
    Average: TAverages;
    AveragedScale: integer;
    Reacted: boolean;
    Distance: single;
  end;

  TNearestSpot = record
    Index: integer;
    Enabled: boolean;
    Name: string;
    LonLat: TPointF;
    ScanPoint: TPoint;
    Distance: single;
    Scale: integer;
  end;

  TNameListRec = record
    Name: string;
    Scale: integer;
  end;

  PAreaRec = ^TAreaRec;
  TAreaRec = record
    Index: integer;
    Code: integer;
    Name: string;
    APoint: TPointF;
    Scale: integer;
    Color: TAlphaColor;
    NameList: TList<TNameListRec>;
  public
    constructor Create(AIndex: integer);
    procedure Free;
    procedure Reset;
    procedure SortNameListByScale;
  end;

  PPsWavesInfo = ^TPsWavesInfo;
  TPsWavesInfo = record
    TravelTime: integer;
    EpicenterF: TPointF;
    Epicenter: TPoint;
    EpicenterToCurrPointF: integer;
    //EpicenterToCurrPoint: integer;
    PWave: TPoint;
    SWave: TPoint;
    PDistance: integer;
    SDistance: integer;
    PRadius: integer;
    SRadius: integer;
    ArrivalSec: integer;
    EpicenterScanned: boolean;
  end;

  TJmaColorScale = record
    Scale: integer;
    AColor: TAlphaColor;
  end;

  TTravelTimeRec = record
    PTime, STime: single;
    Depth: integer;
    Distance: integer;
  end;

  TAccessKind = (akEEW, akKyoshin, akEstInt, akPSWaves);

  PRequestRec = ^TRequestRec;
  TRequestRec = record
    DateTime: TDateTime;
    URL: string;
    AccessKind: TAccessKind;
    KyoshinIndex: TKyoshinIndex;
  end;

  PResponseRec = ^TResponseRec;
  TResponseRec = record
    DateTime: TDateTime;
    AccessKind: TAccessKind;
    Stream: TMemoryStream;
    KyoshinIndex: TKyoshinIndex;
  end;

  TMapDrawRect = (mrLeftTop, mrRightTop, mrLeftBottom, mrRightBottom, mrAll);

  {$IFNDEF MSWINDOWS}
  TFMXSystemFontServiceHook = class(TInterfacedObject, IFMXSystemFontService)
  private var
    FOrgFMXSystemFontService: IFMXSystemFontService;
  protected
    constructor Create;
  public
    function GetDefaultFontFamilyName: String;
    function GetDefaultFontSize: Single; inline;
  public
    class procedure RegistService;
  end;
  {$ENDIF}



  procedure InitFolder;

  function Clock: TDateTime;

  procedure SetMapParams;
  function GetTextWidth(Text: string; AFont: TFont): single;
  function GetTextRect(Text: string; AFont: TFont): TRectF;
  function AdjustLabelFontSize(Text: string; AFont: TFont; AWidth: single): single;

  function GetSpotList: TList;
  procedure ClearSpotList(List: TList);
  function GetAreaList: TList;
  procedure ClearAreaList(List: TList);
  procedure ResetAreaList(List: TList);
  procedure SortAreaList(List: TList);

  function CreateHttpClient: THTTPClient;
  function GetTextFromSite(URL: string): string;
  function GetStreamFromSite(URL: string): TMemoryStream;
  function GetBitmapFromSite(URL: string): TBitmap;

  function TryStrToInteger(Text: string): integer;
  function TryStrToSingle(Text: string): single;

  function LoadBitmapFromResource(AName: string): TBitmap;
  function LoadStreamFromResource(AName: string): TMemoryStream;
  function LoadTextFromResource(AName: string): string;
  procedure CopyBitmapToClipboard(ABitmap: TBitmap);
  procedure SetTextToClipborad(Text: string);

  function GetHubenyDistance(Point1, Point2: TPointF): single;
  function LatLonToStr(APoint: TPointF): string;
  function PointToStr(APoint: TPoint): string;
  function Get_Ex_MapPos(APoint: TPointF): TPoint;
  function Get_Ex_LonLan(APoint: TPoint): TPointF;
  function MousePosToExMapPos(APoint: TPoint): TPoint;
  function MapPosToExMapPos(APoint: TPoint): TPoint;

  function PopFromQueue(Queue: TThreadList): pointer;
  procedure PushToQueue(Queue: TThreadList; P: Pointer);
  function GetCountQueue(Queue: TThreadList): integer;

  function IsGray(Color: TAlphaColor): boolean;
  function ColorToIndexScale(Color: TAlphaColor): integer;
  function GetScaleValueFromHSV(AColor: TAlphaColor): integer;
  function GetJmaIntFromHSV(AColor: TAlphaColor): single;
  function GetTextColor(BackColor: TAlphaColor): TAlphaColor;

  function StrToDateTimeEx(S: string): TDateTime;
  function StrToDateTimeEx2(S: string): TDateTime;

  function GetJmaIntByScale(Scale: integer): string;
  function GetShindoStrByScale(Scale: integer): string;
  function GetJmaIntColorByScale(Scale: integer; Interpolation: boolean = false): TAlphaColor;
  function GetColorByJmaInt(JmaInt: string): TAlphaColor;
  function JmaIntToScale(kaikyu: string): integer;
  function ColorToScale(AColor: TAlphaColor): single;

  procedure LogWriteToFile(Rec: PLogRec);
  procedure AddToLog(LogType: TLogType; Message1, Message2: string); overload;
  procedure AddToLog(Message1, Message2: string); overload;


  // c[p
  procedure SecondToDistance(TravelTime: single; Depth: integer; var PDistance, SDistance: single);
  function GetEstShindo(Epicenter: TPointF; Depth: integer; Magnitude: single;
            CurrentPoint: TPointF; AmpFactor: single): single;
  function GetMeshInfo(APoint: TPointF; v: integer): string;
  procedure LoadTravelTimeTable;


var
  LogQueue: TThreadList = nil;
  ColorBlockArray: array[0..7, 0..7, 0..7] of TList<Integer>;

  AppName: string = '';
  AppDir: string;
  CaptureDir: string;

const
  MapSize: TSize = (cx:352; cy:400);
  Ex_MapSize: TSize = (cx:476; cy:557);

  BooleanStr: array[Boolean] of string = ('No', 'Yes');

  // J[e[u
  KyoshinAlphaColors: array[0..99] of TAlphaColor =
      ($FF0003CF, $FF000BD4, $FF0010D7, $FF0014DA, $FF001ADD, $FF0022E2, $FF002CE9, $FF0033ED,  // 0
       $FF003CF2, $FF0041F6, $FF004BF8, $FF005BEA, $FF0068E0, $FF0071D8, $FF0082CA, $FF008CC2,  // 8
       $FF009EB4, $FF00ACA9, $FF00B3A2, $FF00C495, $FF02D189, $FF08D580, $FF0CD87A, $FF12DC72,  // 16
       $FF19E069, $FF22E65D, $FF27EA56, $FF2DEE4D, $FF34F244, $FF3AF73D, $FF44FA34, $FF50FB30,  // 24
       $FF5BFB2C, $FF64FB2A, $FF74FC24, $FF82FD20, $FF8CFD1C, $FF9DFE17, $FFA3FE14, $FFB2FF10,  // 32
       $FFBBFF0D, $FFC3FF0A, $FFCCFF09, $FFD3FF08, $FFDBFF06, $FFDBFF06, $FFE3FF05, $FFEBFF03,  // 40
       $FFF2FF02, $FFFDFC00, $FFFDFC00, $FFFDFC00, $FFFDFC00, $FFFFF500, $FFFFF200, $FFFFED00,  // 48
       $FFFFE900, $FFFFE500, $FFFFE200, $FFFFDE00, $FFFFD900, $FFFFD200, $FFFFCE00, $FFFFC500,  // 56
       $FFFFBD00, $FFFFB200, $FFFFAD00, $FFFFA600, $FFFF9E00, $FFFF9600, $FFFF8D00, $FFFF8500,
       $FFFF7E00, $FFFF7900, $FFFF6D00, $FFFF6600, $FFFF5D00, $FFFF5900, $FFFF5100, $FFFF4900,
       $FFFF4100, $FFFD3900, $FFFD3500, $FFFC2E00, $FFFB2500, $FFFA1E00, $FFF91900, $FFF81100,
       $FFF80D00, $FFF70500, $FFF10000, $FFE90000, $FFE50000, $FFDD0000, $FFD60000, $FFCC0000,
       $FFC50000, $FFBD0000, $FFB50000, $FFAF0000);

  // o䗦̐F: 4F=98.41935844 %, 8F=99.53108469 %
  HighAppearColor: array[0..7, 0..1] of Cardinal = (
     ($FF0000CD, 0), ($FF00CF8B, 20), ($FF1EF453, 26), ($FF35F840, 29),
     ($FF4CFB2E, 31), ($FF63FF1C, 33), ($FF74FE18, 34), ($FF86FE15, 36));

  KyoshinIndexURLStr: array[TKyoshinIndex] of string =
      ('acmap_s', 'jma_s', 'vcmap_s', 'dcmap_s',
       'rsp0125_s', 'rsp0250_s', 'rsp0500_s', 'rsp1000_s',
       'acmap_b', 'jma_b', 'vcmap_b', 'dcmap_b',
       'rsp0125_b', 'rsp0250_b','rsp0500_b', 'rsp1000_b');

  KyoshinIndexDisplayStr: array[TKyoshinIndex] of string =
      ('n\őx', 'n\A^Ckx', 'n\ő呬x', 'n\őψ',
       'n\0.125Hzx', 'n\0.25Hzx', 'n\0.5Hzx', 'n\1.0Hzx',
       'nőx', 'nA^Ckx', 'nő呬x', 'nőψ',
       'n0.125Hzx', 'n0.25Hzx', 'n0.5Hzx', 'n1.0Hzx');

  LogFilename: array[TLogType] of string = ('Error.log', 'Info.log', 'EEW.log', 'NetError.log',
      'Alert.log', 'NTP.log', 'Debug.log');

  // CےkxK̐F
  // https://www.jma.go.jp/jma/kishou/info/colorguide/HPColorGuide_202007.pdf
  JmaIntColor: array[0..9] of TAlphaColor = ($FFFFFFFF, $FFf2f2ff, $FF00aaff, $FF0041ff, $FFfae696,
                                             $FFffe600, $FFff9900, $FFff2800, $FFa50021, $FFb40068);
  // add 2023-07-20
  JmaIntColorScale: array[0..9] of TJmaColorScale = (
    (Scale:30; AColor:$FFFFFFFF), (Scale:35; AColor:$FFf2f2ff), (Scale:45; AColor:$FF00aaff),
    (Scale:55; AColor:$FF00FF50{FF0041ff}),{ change to $FF00FF50 ?} (Scale:65; AColor:$FFfae696), (Scale:75; AColor:$FFffe600),
    (Scale:80; AColor:$FFff9900), (Scale:85; AColor:$FFff2800), (Scale:90; AColor:$FFa50021),
    (Scale:95; AColor:$FFb40068));


  TextColorWB: array[boolean] of TAlphaColor = (TAlphaColorRec.Black, TAlphaColorRec.White);

  MagnitudeColor: array[0..10] of TAlphaCOlor = (
    $FFC3C3C3, $FF525252, $FF0000FF, $FF008B3C, $FFFFA500,
    $FFF3670A, $FFFF0000, $FFBA00FF, $FFBA00FF, $FFBA00FF, $FFBA00FF);


implementation

uses
  MainUnit, KyoshinUnit;

var
  // c[p
  TravelTimeArray: array[0..15] of TArray<TTravelTimeRec>;

  // Wϊp̃p[^
  Ex_X_PixelsPerDegree: single;
  Ex_Y_PixelsPerDegree: single;
  Ex_X_DegreesPerPixel: single;
  Ex_Y_DegreesPerPixel: single;


const
  // kj^̊gn}͈̔́ioxܓxj
  Ex_StartPoint: TPointF = (X:122.51; Y:23.6);
  Ex_EndPoint: TPointF = (X:145.9; Y:46.3);

  MinPWavesSec: array[0..14] of integer = (0, 2, 4, 4, 7, 8, 9, 11, 12, 13, 14, 16, 17, 18, 19);
  MinSWavesSec: array[0..14] of integer = (0, 4, 6, 6, 11, 14, 16, 18, 20, 23, 25, 27, 29, 32, 34);



// misc functions

procedure InitFolder;
begin
  try
    {$IFDEF MSWINDOWS}
    AppDir := ExtractFileDir(ParamStr(0));
    AppName := TPath.GetFileNameWithoutExtension(ParamStr(0));
    {$ENDIF}

    {$IFDEF MACOS}
    // macOSł"~/Library/Application Support" tH_
    // %MyApp% tH_ɃOt@CLv`摜ۑ
    AppName := TPath.GetFileNameWithoutExtension(ParamStr(0));
    if not AppName.IsEmpty then
      AppDir := TPath.GetLibraryPath + '/Application Support/' + AppName
    else
      AppDir := TPath.GetLibraryPath + '/Application Support/Kyoshin';
    if not TDirectory.Exists(AppDir) then begin
      TDirectory.CreateDirectory(AppDir);
    end;
    {$ENDIF}

    {$IFDEF ANDROID}
    // OуLv`摜̕ۑtH_
    AppDir := TPath.GetPublicPath;
    // "/storage/emulated/0/Android/data/info.melanion.%AppName%/files"
    var SArray := SplitString(AppDir, './');
    if Length(SArray) >= 2 then
      AppName := SArray[Length(SArray) - 2];
    AppDir := TPath.Combine(AppDir, 'AppHome');
    if not TDirectory.Exists(AppDir) then
      TDirectory.CreateDirectory(AppDir);
    {$ENDIF}

    CaptureDir := AppDir + PathDelim + 'Capture';
    if not TDirectory.Exists(CaptureDir) then
      TDirectory.CreateDirectory(CaptureDir);

  except
    on e: exception do
      AddToLog('InitFolder', e.Message);
  end;
end;


// Wϊp[^̃Zbg
procedure SetMapParams;
begin
  // 1xpx
  Ex_X_PixelsPerDegree := Ex_MapSize.cx / (Ex_EndPoint.X - Ex_StartPoint.X);
  Ex_Y_PixelsPerDegree := Ex_MapSize.cy / (Ex_EndPoint.Y - Ex_StartPoint.Y);

  // 1px̓x
  Ex_X_DegreesPerPixel := (Ex_EndPoint.X - Ex_StartPoint.X) / Ex_MapSize.cx;
  Ex_Y_DegreesPerPixel := (Ex_EndPoint.Y - Ex_StartPoint.Y) / Ex_MapSize.cy;
end;

// oxܓxgn}̍Wɕϊ
function Get_Ex_MapPos(APoint: TPointF): TPoint;
begin
  result.X := Trunc((APoint.X - Ex_StartPoint.X) / Ex_X_DegreesPerPixel);
  result.Y := Trunc(Ex_MapSize.cy - 1 - (APoint.Y - Ex_StartPoint.Y) / Ex_Y_DegreesPerPixel);
end;

// gn}̈ʒuoxܓxɕϊ
function Get_Ex_LonLan(APoint: TPoint): TPointF;
begin
  result.X := Ex_StartPoint.X + APoint.X / Ex_X_PixelsPerDegree;
  result.Y := Ex_StartPoint.Y + (Ex_MapSize.cy - APoint.Y) / Ex_Y_PixelsPerDegree;
end;

// \ꂽ摜̈ʒugn}̈ʒuɕϊ
function MousePosToExMapPos(APoint: TPoint): TPoint;
begin
  result.X := DisplayedMapRect.Left + Round(APoint.X * MapSize.cx / MainForm.ClientWidth);
  result.Y := DisplayedMapRect.Top + Round(APoint.Y * MapSize.cy / MainForm.ClientHeight);
end;

// kj^̃IWin}̈ʒugn}̈ʒuɕϊ
function MapPosToExMapPos(APoint: TPoint): TPoint;
begin
  result.X := (Ex_MapSize.cx - MapSize.cx) + Round(APoint.X * MapSize.cx / MainForm.PaintBox1.Width);
  result.Y := Round(APoint.Y * MapSize.cy / MainForm.PaintBox1.Height);
end;


function LonLanStrToPointF(LonLanStr: string): TPointF;
var
  SArray: TStringDynArray;
begin
  result := TPointF.Zero;
  SArray := SplitString(LonLanStr, ',');  // ܓx ox
  SArray[0] := Trim(SArray[0]);
  if Length(SArray) = 2 then begin
    if not TryStrToFloat(SArray[0], result.Y) then
      TryStrToFloat(Copy(SArray[0], 1, Length(SArray[0]) - 1), result.Y);
    if not TryStrToFloat(SArray[1], result.X) then
      TryStrToFloat(Copy(SArray[1], 1, Length(SArray[1]) - 1), result.X);

  end;
end;

function LatLonToStr(APoint: TPointF): string;
begin
  result := '';
  if not APoint.IsZero then
    result := APoint.Y.ToString(TFloatFormat.ffFixed, 7, 2) + 'N ' +
              APoint.X.ToString(TFloatFormat.ffFixed, 7, 2) + 'E';
end;

function PointToStr(APoint: TPoint): string;
begin
  result := APoint.X.ToString + ', ' + APoint.Y.ToString;
end;

function GetTextWidth(Text: string; AFont: TFont): single;
var
  Layout: TTextLayout;
begin
  result := 0;

  Layout := TTextLayoutManager.DefaultTextLayout.Create;
  try
    Layout.BeginUpdate;
    try
      Layout.Font.Assign(AFont);
      Layout.HorizontalAlign := TTextAlign.Leading;
      Layout.WordWrap := false;
      Layout.Text := Text;
    finally
      Layout.EndUpdate;
    end;
    result := Layout.TextWidth;
  finally
    Layout.Free;
  end;
end;

function GetTextRect(Text: string; AFont: TFont): TRectF;
var
  Layout: TTextLayout;
begin
  result := TRectF.Empty;

  Layout := TTextLayoutManager.DefaultTextLayout.Create;
  try
    Layout.BeginUpdate;
    try
      Layout.Font.Assign(AFont);
      Layout.HorizontalAlign := TTextAlign.Leading;
      Layout.VerticalAlign := TTextAlign.Leading;
      Layout.WordWrap := false;
      Layout.Text := Text;
    finally
      Layout.EndUpdate;
    end;
    result := Layout.TextRect;
  finally
    Layout.Free;
  end;
end;

function AdjustLabelFontSize(Text: string; AFont: TFont; AWidth: single): single;
var
  Layout: TTextLayout;
begin
    Layout := TTextLayoutManager.DefaultTextLayout.Create;
    try
      Layout.BeginUpdate;
      try
        Layout.Font.Assign(AFont);
        Layout.Font.Size := 32;
        Layout.HorizontalAlign := TTextAlign.Leading;
        Layout.WordWrap := false;
        Layout.Text := Text;

      finally
        Layout.EndUpdate;
      end;
      while Layout.TextWidth > AWidth do
        Layout.Font.Size := Layout.Font.Size - 0.5;
      result := Layout.Font.Size;
    finally
      Layout.Free;
    end;
end;

// Ȍ
procedure LogWriteToFile(Rec: PLogRec);
var
  TheTime: TDateTime;
begin
  try
    if Rec <> nil then begin
      if Rec.DateTime = 0 then
        TheTime := Clock
      else
        TheTime := Rec.DateTime;

      if (Rec.LogType <= TLogType.ltError) and (Rec.LogType <= TLogType.ltDebug) then begin
        if Rec.Message1 = '-' then begin
          TFile.AppendAllText(AppDir + PathDelim + LogFileName[Rec.LogType],
                       DupeString('-', 13) + sLineBreak + Rec.Message2,
                       TEncoding.UTF8);
        end else if Rec.Message2 <> '' then
          TFile.AppendAllText(AppDir + PathDelim + LogFileName[Rec.LogType],
                       FormatDateTime('yyyy/mm/dd hh:nn:ss-', TheTime ) +
                       Rec.Message1 + ':' + Rec^.Message2 + sLineBreak,
                       TEncoding.UTF8)
        else
          TFile.AppendAllText(AppDir + PathDelim + LogFileName[Rec.LogType],
                       FormatDateTime('yyyy/mm/dd hh:nn:ss-', TheTime ) +
                       Rec.Message1 + sLineBreak,
                       TEncoding.UTF8);
      end;
    end;
  except

  end;
end;

procedure LogWriteDirect(LogType: TLogType; Text: string);
begin
  try
    if not Text.IsEmpty and (LogType <= TLogType.ltError) and (LogType <= TLogType.ltDebug) then
      TFile.AppendAllText(AppDir + PathDelim + LogFileName[LogType],
                   FormatDateTime('yyyy/mm/dd hh:nn:ss-', Clock) +
                   Text + sLineBreak,
                   TEncoding.UTF8);
  except

  end;
end;

procedure AddToLog(LogType: TLogType; Message1, Message2: string);
var
  Rec: PLogRec;
begin
  if AppClosing or not Assigned(LogQueue) then
    LogWriteDirect(LogType, Message1 + '-' + Message2)
  else
    try
      New(Rec);
      Rec.DateTime := Clock;
      Rec.LogType := LogType;
      Rec.Message1 := Message1;
      Rec.Message2 := Message2;
      PushToQueue(LogQueue, Rec)
    except

    end;
end;

procedure AddToLog(Message1, Message2: string);
begin
    AddToLog(TLogType.ltError, Message1, Message2);
end;

// NTPŒ    kj^̎Ԃׂ͂Ăg
function Clock: TDateTime;
begin
  result := Now + NTPTimeDiffer;
end;


function TryStrToInteger(Text: string): integer;
begin
  if not TryStrToInt(Text, result) then
    result := 0;
end;

function TryStrToSingle(Text: string): single;
begin
  if not TryStrToFloat(Text, result) then
    result := 0;
end;

// L[
procedure PushToQueue(Queue: TThreadList; P: Pointer);
begin
  if (Queue <> nil) and (P <> nil) then begin
    try
      with Queue.LockList do begin
        Add(P);
      end;
    finally
      Queue.UnlockList;
    end;
  end;
end;

function PopFromQueue(Queue: TThreadList): pointer;
begin
  result := nil;

  if Queue <> nil then begin
    try
      with Queue.LockList do begin
        if Count > 0 then begin
          result := Items[0];
          Delete(0);
        end;
      end;
    finally
      Queue.UnlockList;
    end;
  end;
end;

function GetCountQueue(Queue: TThreadList): integer;
begin
  result := 0;

  if Queue <> nil then begin
    try
      with Queue.LockList do begin
        result := Count;
      end;
    finally
      Queue.UnlockList;
    end;
  end;
end;

//
// J[֌W
//

function IsGray(Color: TAlphaColor): boolean;
begin
    result := (TAlphaColorRec(Color).R = TAlphaColorRec(Color).G) and
              (TAlphaColorRec(Color).G = TAlphaColorRec(Color).B);
end;

// --- J[e[uT iŋߖT_Tj

// kj^摜瓾ꂽϑ_̐FƃJ[e[uFƂRGB32ȓƂO
// RGBԂ8x8x8̃ubNɕAubNColorBlockArrayzŊǗ
// eubÑXgɂ́AYubNїאڂ26̃ubɑ̃J[e[uFւ̃CfbNX܂܂Ă
// Xg̑SJ[e[uFTARGB32ȓŁAƂRGB߂J[e[uFŋߎFƂȂ
function ColorToIndexScale(Color: TAlphaColor): integer;
var
  R, G, B: integer;
  AColor0, AColor1: TAlphaColorRec;
  NearestIndex: integer;
  Value, NearestValue: single;
begin
  result := -1;

  try
    // ---- o̐F̌
    for var i := 0 to 7 do
      if Color = HighAppearColor[i, 0] then begin
        result := HighAppearColor[i, 1];
        Exit;         // T
      end;

    // ---- ߖT̃ubNT   8x8x8̃ubN̒ߖT27̃ubNT
    NearestIndex := $FF;
    NearestValue := MaxSingle;
    AColor0.Color := Color;
    R :=  AColor0.R SHR 5;  // div 32
    G :=  AColor0.G SHR 5;
    B :=  AColor0.B SHR 5;
    if ColorBlockArray[R, G, B] <> nil then begin
      // ʏ8x̃[v
      for var i in ColorBlockArray[R, G, B] do begin   // TΏۂ̐FubÑXggĒT
        AColor1.Color := KyoshinAlphaColors[i];
        // [NbhvZ
        Value := SQRT(Sqr(AColor0.R - AColor1.R) + Sqr(AColor0.G - AColor1.G) + Sqr(AColor0.B - AColor1.B));
        if Value < NearestValue then begin
          NearestValue := Value;
          NearestIndex := i;
        end;
      end;
    end;
    // RGB32ȓ̂ݗL
    if (NearestIndex <> $FF) and (NearestValue <= 32) then begin
      result := NearestIndex;
      exit;           // T
    end;

    // J[ubNTsƂɃCNg
    TInterlocked.Increment(ColorSearch2Count);

    // ---- 2@̒TɎsƂ͑S100FT  -  s͂ȂÔ
    NearestIndex := $FF;
    NearestValue := MaxSingle;
    AColor0.Color := Color;
    for var i := 0 to 99 do begin
      AColor1.Color := KyoshinAlphaColors[i];
      Value := Sqr(AColor0.R - AColor1.R) + Sqr(AColor0.G - AColor1.G) + Sqr(AColor0.B - AColor1.B);
      if Value < NearestValue then begin
        NearestValue := Value;
        NearestIndex := i;
      end;
    end;
    result := NearestIndex;

    if result = $FF then
      result := -1;
  except
    on e: exception do
      AddToLog('ColorToIndexScale', e.Message);
  end;
end;

// ԂɂFElϊ
// https://qiita.com/NoneType1/items/a4d2cf932e20b56ca444
function ColorToScale(AColor: TAlphaColor): single;
  type
    THSVRec = record
      H, S, V: single;
    end;
  function RGBtoHSV(AColor: TAlphaColor):THSVRec;
  var
    RGBMax, RGBMin: Byte;
    AColorRec: TAlphaColorRec;
  begin
    try
      AColorRec.Color := AColor;

      RGBMax := Max(AColorRec.R, AColorRec.G);
      RGBMax := Max(RGBMax, AColorRec.B);
      RGBMin := Min(AColorRec.R, AColorRec.G);
      RGBMin := Min(RGBMin, AColorRec.B);
      // Hue 0`1
      if RGBMax = RGBMin then
        result.H := 0
      else if AColorRec.R = RGBMax then
        result.H := 60 * (AColorRec.G - AColorRec.B) / (RGBMax - RGBMin)
      else if AColorRec.G = RGBMax then
        result.H := 60 * (AColorRec.B - AColorRec.R) / (RGBMax - RGBMin) + 120
      else
        result.H := 60 * (AColorRec.R - AColorRec.G) / (RGBMax - RGBMin) + 240;
      result.H := result.H / 360;

      result.H := FMod(result.H, 360);
      if result.H < 0 then
        result.H := result.H + 360;
      // Saturation or Chroma 0`1
      if RGBMax = 0 then
        result.S := 0
      else
        result.S := (RGBMax - RGBMin) / RGBMax;
      // Value or Brightness 0`1
      result.V := RGBMax / 255;
    except

    end;
  end;
var
  HSV: THSVRec;
begin
 result := 0;

  HSV := RGBtoHSV(AColor);

  try
    if (HSV.v > 0.1) and (HSV.s > 0.75) then begin
        if HSV.h > 0.1476 then
            result := 280.31 * Power(HSV.h, 6) - 916.05 * Power(HSV.h, 5) + 1142.6 * Power(HSV.h, 4) -
                      709.95 * Power(HSV.h,3) + 234.65 * Power(HSV.h, 2) - 40.27 * HSV.h + 3.2217;

        if (HSV.h <= 0.1476) and  (HSV.h > 0.001) then
            result := 151.4 * Power(HSV.h,4) -49.32 * Power(HSV.h,3) + 6.753 * Power(HSV.h,2) -2.481 * HSV.h + 0.9033;

        if HSV.h <= 0.001 then
            result :=   -0.005171 * Power(HSV.v,2) - 0.3282 * HSV.v + 1.2236;
    end;

    if result < 0 then
        result := 0;
  except

  end;
end;

function GetScaleValueFromHSV(AColor: TAlphaColor): integer;
begin
  result := Round(ColorToScale(AColor) * 100);
end;

function GetJmaIntFromHSV(AColor: TAlphaColor): single;
begin
  result := 10 * ColorToScale(AColor) - 3;
end;

function GetPGAFromHSV(AColor: TAlphaColor): single;
begin
  result := 105 * ColorToScale(AColor) - 2;
end;

function GetPGVFromHSV(AColor: TAlphaColor): single;
begin
  result := 105 * ColorToScale(AColor) - 3;
end;

function GetTextColor(BackColor: TAlphaColor): TAlphaColor;
var
  AColorRec: TAlphaColorRec;
begin
  AColorRec.Color := BackColor;
  result := TextColorWB[((AColorRec.R * 299) + (AColorRec.G * 587) + (AColorRec.B * 114)) / 1000 < 128]
end;



function ScaleToIndex(Scale: integer): integer;
const
  AIndex: array[0..19] of integer = (
    -1,  -1,  -1,  -1,  -1,  -1,   0,   1,   1,   2,   2,   3,   3,   4,   4,   5,   6,   7,   8,   9);
begin
  if InRange(Scale, 0, 99) then
    result := AIndex[Trunc(Scale / 5)]
  else
    result := -1;

end;

// CےkxK
// https://www.data.jma.go.jp/eqev/data/kyoshin/kaisetsu/calc_sindo.html
function GetJmaIntByScale(Scale: integer): string;
const
  IndexStr: array[0..9] of string = ('0', '1', '2', '3', '4', '5', '5', '6', '6', '7');
var
  AIndex: integer;
begin
  result := '';

  AIndex := ScaleToIndex(Scale);
  if InRange(AIndex, Low(IndexStr), High(IndexStr)) then
    result := IndexStr[AIndex]
  else
    result := '-';
end;

function JmaIntToSingle(kaikyu: string): single;
begin
  if not TryStrToFloat(Kaikyu, result) then begin
    if (Kaikyu = '6') or (Kaikyu = '6+') then
      result := 6
    else if (Kaikyu = '6') or (Kaikyu = '6-') then
      result := 5.5
    else if (Kaikyu = '5') or (Kaikyu = '5+') then
      result := 5
    else if (Kaikyu = '5') or (Kaikyu = '5-') then
      result := 4.5
    else
      result := 0;
  end;
end;

function JmaIntToScale(kaikyu: string): integer;
begin
  result := EnsureRange(Round(JmaIntToSingle(kaikyu) * 10 + 30), 0, 99);
end;

function GetColorByJmaInt(JmaInt: string): TAlphaColor;
var
  ColorIndex: integer;
begin
  result := TAlphaColorRec.White;
  try
    if JmaInt = '7' then
      ColorIndex := 9
    else if (JmaInt = '6') or (JmaInt = '6+') then
      ColorIndex := 8
    else if (JmaInt = '6') or (JmaInt = '6-') then
      ColorIndex := 7
    else if (JmaInt = '5') or (JmaInt = '5+') then
      ColorIndex := 6
    else if (JmaInt = '5') or (JmaInt = '5-') then
      ColorIndex := 5
    else
      ColorIndex := TryStrToInteger(JmaInt);
    result := JmaIntColor[ColorIndex];
  except

  end;
end;

// u݂̐kxv̐kxC[Wp̕
function GetShindoStrByScale(Scale: integer): string;
const
  IndexStr: array[0..9] of string = ('0', '1', '2', '3', '4', '5-', '5+', '6-', '6+', '7');
var
  AIndex: integer;
begin
  result := '';

  AIndex := ScaleToIndex(Scale);
  if InRange(AIndex, Low(IndexStr), High(IndexStr)) then
    result := IndexStr[AIndex];
end;

function GetJmaIntColorIndexToScale(AIndex: integer): integer;
const
  Scale: array[0..9] of integer = (30, 35, 45, 55, 65, 75, 80, 85, 90, 95);
begin
  if InRange(AIndex, 0, 9) then
    result := Scale[AIndex]
  else
    result := -1;
end;

// ScalelCےkxKFɕϊ
// https://www.jma.go.jp/jma/kishou/info/colorguide/HPColorGuide_202007.pdf
function GetJmaIntColorByScale(Scale: integer; Interpolation: boolean = false): TAlphaColor;
var
  AIndex: integer;
  AColor1, AColor2, resultColor: TAlphaColorRec;
  Rec1, Rec2: TJmaColorScale;
begin
  result := TAlphaColorRec.White;

  if inRange(Scale, 30, 99) then begin
    try
      if not Interpolation or (Scale >= 95) then begin
         result := JmaIntColor[EnsureRange(ScaleToIndex(Scale), 0, 9)];
      end else begin
        // 
        // alter 2023-07-20
        AIndex := ScaleToIndex(Scale);
        Rec1 := JmaIntColorScale[AIndex];
        Inc(AIndex);
        Rec2 := JmaIntColorScale[AIndex];
        AColor1.Color := Rec1.AColor;
        AColor2.Color := Rec2.AColor;

        resultColor.R := AColor1.R + Trunc((AColor2.R - AColor1.R) / (Rec2.Scale - Rec1.Scale) * (Scale - Rec1.Scale));
        resultColor.G := AColor1.G + Trunc((AColor2.G - AColor1.G) / (Rec2.Scale - Rec1.Scale) * (Scale - Rec1.Scale));
        resultColor.B := AColor1.B + Trunc((AColor2.B - AColor1.B) / (Rec2.Scale - Rec1.Scale) * (Scale - Rec1.Scale));
        resultColor.A := $FF;

        result := resultColor.Color;
      end;
    except

    end;
  end;
end;



// qxǰɂ2_̌oxEܓxin\ʁj擾 km
function GetHubenyDistance(Point1, Point2: TPointF): single;
var
  Lon1Rad, Lat1Rad, Lon2Rad, Lat2Rad: single;
  LatAveRad, LatDiff, LonDiff: single;
  Meridian, PrimeVertical: single;
  X, Y: single;
begin
  result := 0;

  try
    //WAɕϊ
    Lon1Rad := Point1.X * PI / 180;
    Lat1Rad := Point1.Y * PI / 180;
    Lon2Rad := Point2.X * PI / 180;
    Lat2Rad := Point2.Y * PI / 180;

    LatAveRad := (Lat1Rad + Lat2Rad) / 2;
    LatDiff := Lat1Rad - Lat2Rad;
    LonDiff := Lon1Rad - Lon2Rad;

    //qߐȗa
    Meridian := 6334834 / Sqrt(Power(1 - 0.006674 * Sin(LatAveRad) * Sin(LatAveRad), 3));
    //Kѐȗa
    PrimeVertical := 6377397 / Sqrt(1 - 0.006674 * Sin(LatAveRad) * Sin(LatAveRad));

    // qxǰ
    X := Meridian * LatDiff;
    Y := PrimeVertical * Cos(LatAveRad) * LonDiff;

    Result := Sqrt(X * X + Y * Y) / 1000;  // m >> km
  except

  end;
end;

//
// ȈՑ\֘A
//

procedure LoadTravelTimeTable;
var
  SList: TStringList;
  SArray: TStringDynArray;
  LIndex: integer;
  RIndex: integer;
  Depth: integer;
  Rec: TTravelTimeRec;
begin
  for var i := 0 to 15 do
    SetLength(TravelTimeArray[i], 236);

  SList := TStringList.Create;
  try
    try
      SList.Text := LoadTextFromResource('TravelTimeTable');
      LIndex := 0;
      RIndex := 0;
      Depth := 0;
      for var Text in SList do begin
        SArray := SplitString(Text, ',');
        if Length(SArray) = 4 then begin
          Rec.PTime := SArray[0].ToSingle;
          Rec.STime := SArray[1].ToSingle;
          Rec.Depth := StrToInt(SArray[2]);
          Rec.Distance := StrToInt(SArray[3]);
          if Depth <> Rec.Depth then begin
            Depth := Rec.Depth;
            Inc(LIndex);
            RIndex := 0;
          end;
          TravelTimeArray[LIndex][RIndex] := Rec;
          Inc(RIndex);
        end;
      end;
    except

    end;
  finally
    SList.Free;
  end;
end;

// ȈՑ\gđP/Sg̋擾
// [10kmPʂ150km܂ŁA 240b܂
procedure SecondToDistance(TravelTime: single; Depth: integer; var PDistance, SDistance: single);
  type
    TParaSingle = array[0..2] of single;
  // 3_OWԁiEEWł͐[10kmPʂȂ̂ŁA3_ԂɊȗj
  function Interpolation(Value: single; Seconds: TParaSingle; Distances: TParaSingle): single;
  var
    sum, prod: single;
    i, j: Integer;
  begin
    sum := 0;
    for i := Low(Seconds) to High(Seconds) do begin
      prod := Distances[i];
      for j := Low(Seconds) to High(Seconds) do begin
        if j <> i then
          prod := prod * (Value - Seconds[j]) / (Seconds[i] - Seconds[j]);
      end;
      sum := sum + prod;
    end;
    result := sum;
  end;
var
  ListIndex: integer;
  TTArray: TArray<TTravelTimeRec>;
  Rec: TTravelTimeRec;
  FoundIndex: integer;
  MinPTime, MinSTime: single;
  Seconds: TParaSingle;
  Distances: TParaSingle;
begin
  PDistance := 0;
  SDistance := 0;

  if TravelTime > 240 then
    Exit;

  ListIndex := Round(Depth / 10);
  if (ListIndex >= 0) and (ListIndex <= 15) then begin
    TTArray := TravelTimeArray[ListIndex];
    Rec := TTArray[0];
    MinPTime := Rec.PTime;
    MinSTime := Rec.STime;
    // p wave
    if MinPTime <= TravelTime then begin
      Rec.PTime := TravelTime;
      // \T
      if TArray.BinarySearch<TTravelTimeRec>(TTArray, Rec, FoundIndex,
        TComparer<TTravelTimeRec>.Construct(
          function (const Left, Right: TTravelTimeRec): Integer
          begin
            result := CompareValue(Left.PTime, Right.PTime);
          end))
                                                                        then begin
         // qbgꍇiۂɂ͂͂Ȃj
         PDistance := TTArray[FoundIndex].Distance;
       end else begin
         // qbgȂꍇATl傫ŏ̃CfbNXFoundIndex ɓĂ
         FoundIndex := FoundIndex - 2;
         if FoundIndex < 0 then
           FoundIndex := 0
         else if FoundIndex > 234 then
           FoundIndex := 234;
         // 3_Zbg
         for var i := 0 to 2 do begin
           Rec := TTArray[FoundIndex + i];
           Seconds[i] := Rec.PTime;
           Distances[i] := Rec.Distance;
         end;
         // 3_ŕ
         PDistance := Interpolation(TravelTime, Seconds, Distances);
       end;
    end;

    // s wave
    if MinSTime <= TravelTime then begin
      Rec.STime := TravelTime;
      if TArray.BinarySearch<TTravelTimeRec>(TTArray, Rec, FoundIndex,
        TComparer<TTravelTimeRec>.Construct(
          function (const Left, Right: TTravelTimeRec): Integer
          begin
            result := CompareValue(Left.STime, Right.STime);
          end))
                                                                        then begin
         SDistance := TTArray[FoundIndex].Distance;
       end else begin
         FoundIndex := FoundIndex - 2;
         if FoundIndex < 0 then
           FoundIndex := 0
         else if FoundIndex > 234 then
           FoundIndex := 234;
         for var i := 0 to 2 do begin
           Rec := TTArray[FoundIndex + i];
           Seconds[i] := Rec.STime;
           Distances[i] := Rec.Distance;
         end;
         SDistance := Interpolation(TravelTime, Seconds, Distances);
       end;
    end;
  end;
end;

const
  EarthRadius = 6370.291;   // n̔aikmj

// kƌݒnƂ̋
function GetDistance(Epicenter: TPointF; Depth: integer; CurrentPoint: TPointF): single;
var
  EqPoint: TPointF;
  CPPoint: TPointF;
  Delta: single;
begin
  result := 0;
  try
    // WAɕϊ
    EqPoint.X := Epicenter.X * PI / 180;     // k
    EqPoint.Y := Epicenter.Y * PI / 180;
    CPPoint.X := CurrentPoint.X * PI / 180;  // ݒn
    CPPoint.Y := CurrentPoint.Y * PI / 180;

    // nܓxnSܓxɕϊ
    EqPoint.Y := EqPoint.Y - (11.55/60) * (PI / 180) * sin(2 * EqPoint.Y);
    CPPoint.Y := CPPoint.Y - (11.55/60) * (PI / 180) * sin(2 * CPPoint.Y);

    // k
    // p5
    Delta := Sqrt(Sqr((cos(EqPoint.Y) * cos(EqPoint.X) * (EarthRadius - Depth) / EarthRadius) -
                      (cos(CPPOint.Y) * cos(CPPOint.X))) +
                  Sqr((cos(EqPoint.Y) * sin(EqPoint.X) * (EarthRadius - Depth) / EarthRadius) -
                      (cos(CPPOint.Y) * sin(CPPOint.X))) +
                  Sqr((sin(EqPoint.Y) * (EarthRadius - Depth) / EarthRadius) - (sin(CPPoint.Y))));

    result := Delta * EarthRadius;
  except
    on e: Exception do
      AddToLog('GetDistance', e.Message);
  end;
end;

function GetDistanceWithMagnitude(Epicenter: TPointF; Depth: integer;  Magnitude: single;
                             CurrentPoint: TPointF): single;
var
  Mag: single;
begin
  result := GetDistance(Epicenter, Depth, CurrentPoint);

  try
    // [g}Oj`[h
    Mag := Magnitude - 0.171;

    // fwŒZ
    result := result - Round(Power(10, 0.5 * Mag - 1.85) / 2);
    if result < 3 then
      result := 3;
  except
    on e: exception do
      AddToLog(ltError, 'GetDistanceWithMagnitude', e.Message);
  end;
end;

// nՑw肵\zkx̌vZ
// http://repository.aitech.ac.jp/dspace/bitstream/11133/2521/1/h4JEً}nk3(P21-26).pdf
function GetEstShindo(Epicenter: TPointF; Depth: integer; Magnitude: single;
               CurrentPoint: TPointF; AmpFactor: single): single;
var
  Mag: single;
  Length: single;

  PGV600, ARV600, PGV: single;
begin
  result := 0;

  try
    // [g}Oj`[h
    Mag := Magnitude  - 0.171;

    Length := GetDistanceWithMagnitude(Epicenter, Depth, Magnitude, CurrentPoint);
    AddToLog(TLogType.ltDebug, 'GetEstShindo', Length.ToString);

    // Տ̍ő呬x
    PGV600 := Power(10, 0.58 * Mag + 0.0038 * Depth -1.29 -
                    log10(Length + 0.0028 * Power(10, 0.50 * Mag)) -0.002 * Length);

    // ϑ_̒nՑ
    ARV600 := AmpFactor;

    // n\ʂł̍ő呬x
    PGV := ARV600 * PGV600;

    // vkx
    result := 2.68 + 1.72 * log10(PGV);
    if result < 0 then
      result := 0;

  except
    on e: exception do
      AddToLog(ltError, 'GetEstShindo', e.Message);
  end;
end;

// nՑ̎擾
function GetMeshInfo(APoint: TPointF; v: integer): string;
var
  URL: string;
  Text: string;
  P0, P1: integer;
begin
  result := '';

  // hЉȌAPI https://www.j-shis.bosai.go.jp/api-sstruct-meshinfo
  if InRange(APoint.Y, 20, 46) and InRange(APoint.X, 122, 154) and InRange(v, 1, 3) then begin
    URL := 'https://www.j-shis.bosai.go.jp/map/api/sstrct/V' + v.ToString + '/meshinfo.gml?' +
           'position=' + APoint.X.ToString(TFloatFormat.ffFixed, 7, 2) + ',' + APoint.Y.ToString(TFloatFormat.ffFixed, 7, 2) +
            '&epsg=4326';
    try
      result := 'error';
      Text := GetTextFromSite(URL);
      // l̎擾
      P0 := Pos('<jshis:ARV>', Text);
      P1 := Pos('</jshis:ARV>', Text);
      if (P0 > 0) and (P1 > 0) then begin
        P0 := P0 + Length('<jshis:ARV>');
        result := Copy(Text, P0, P1 - P0);
      end;
    except
      on e: exception do
        AddToLog('GetMeshInfo', e.Message);
    end;
  end;
end;


// 20110430020430
function StrToDateTimeEx(S: string): TDateTime;
begin
  result := 0;
  if (S <> '') and (Length(S) = 14) then begin
    try
      if not TryEncodeDateTime(StrToInt(Copy(S, 1, 4)),  // year
                               StrToInt(Copy(S, 5, 2)),  // Month
                               StrToInt(Copy(S, 7, 2)),  // Day
                               StrToInt(Copy(S, 9, 2)),  // Hour
                               StrToInt(Copy(S, 11, 2)), // Minute
                               StrToInt(Copy(S, 13, 2)), // Second
                               0, result) then
        result := 0;
    except

    end;
  end;
end;

// 2011/04/30 02:04:30
// 2011/04/30 2:04:12
function StrToDateTimeEx2(S: string): TDateTime;
begin
  result := 0;

  if S.IsEmpty then exit;

  if not TryStrToDateTime(S, result) then
    try
      if Length(S) = 19 then begin
        if not TryEncodeDateTime(StrToInt(Copy(S, 1, 4)),  // year
                                 StrToInt(Copy(S, 6, 2)),  // Month
                                 StrToInt(Copy(S, 9, 2)),  // Day
                                 StrToInt(Copy(S, 12, 2)),  // Hour
                                 StrToInt(Copy(S, 15, 2)), // Minute
                                 StrToInt(Copy(S, 18, 2)), // Second
                                 0, result) then
          result := 0;
      end else if Length(S) = 18 then begin
        if not TryEncodeDateTime(StrToInt(Copy(S, 1, 4)),  // year
                                 StrToInt(Copy(S, 6, 2)),  // Month
                                 StrToInt(Copy(S, 9, 2)),  // Day
                                 StrToInt(Copy(S, 12, 1)),  // Hour
                                 StrToInt(Copy(S, 14, 2)), // Minute
                                 StrToInt(Copy(S, 17, 2)), // Second
                                 0, result) then
          result := 0;
      end;
    except

    end;
end;

// \[X֌W
function LoadBitmapFromResource(AName: string): TBitmap;
var
  Stream: TResourceStream;
begin
  result := nil;
  try
    Stream := TResourceStream.Create(HInstance, AName, RT_RCDATA);
    if Stream <> nil then
      try
        if Stream.Size > 0 then begin
          result := Tbitmap.Create;
          result.LoadFromStream(Stream);
        end;
      finally
        Stream.Free;
      end;
  except

  end;
end;

function LoadStreamFromResource(AName: string): TMemoryStream;
var
  Stream: TResourceStream;
begin
  result := nil;
  try
    Stream := TResourceStream.Create(HInstance, AName, RT_RCDATA);
    if Stream <> nil then begin
      try
        if Stream.Size > 0 then begin
          Stream.Seek(0, TSeekOrigin.soBeginning);
          result := TMemoryStream.Create;
          result.LoadFromStream(Stream);
          result.Seek(0, TSeekOrigin.soBeginning);
        end;
      finally
        Stream.Free;
      end;
    end;
  except

  end;
end;

function LoadTextFromResource(AName: string): string;
var
  Stream: TResourceStream;
  SList: TStringList;
begin
  result := '';

  try
    Stream := TResourceStream.Create(HInstance, AName, RT_RCDATA);
    if Stream <> nil then begin
      try
        if Stream.Size > 0 then begin
          Stream.Seek(0, TSeekOrigin.soBeginning);
          SList := TStringList.Create;
          try
            SList.LoadFromStream(Stream, TEncoding.UTF8);
            result := SList.Text;
          finally
            SList.Free;
          end;
        end;
      finally
        Stream.Free;
      end;
    end;
  except

  end;
end;

// Nbv{[hɉ摜Rs[
procedure CopyBitmapToClipboard(ABitmap: TBitmap);
var
  Clipboard: IFMXExtendedClipboardService;
  BitmapSurface: TBitmapSurface;
begin
  if TPlatformServices.Current.SupportsPlatformService(IFMXExtendedClipboardService, Clipboard) then begin
    BitmapSurface := TBitmapSurface.Create;
    try
      try
        if ABitmap <> nil then begin
          BitmapSurface.Assign(ABitmap);
          Clipboard.SetImage(BitmapSurface);
        end;
      except

      end;
    finally
      BitmapSurface.Free;
    end;
  end;

end;

procedure SetTextToClipborad(Text: string);
var
  Clipboard: IFMXClipboardService;
begin
  if TPlatformServices.Current.SupportsPlatformService(IFMXClipboardService, Clipboard) then
    Clipboard.SetClipboard(Text);
end;

// lbg֌W

function CreateHttpClient: THTTPClient;
begin
  result := THTTPClient.Create;

  result.AllowCookies := true;
  result.MaxRedirects := 10;
  result.HandleRedirects := true;

  result.Accept := 'text/html\tapplication/xhtml+xml\tapplication/xml;q=0.9\timage/webp\timage/apng\t*/*;q=0.8\tapplication/signed-exchange;v=b3;q=0.9';
  result.AcceptLanguage := 'ja\ten-US;q=0.9\ten;q=0.8';
  result.UserAgent := 'Kyoshiin Monitor test';

  {$IFDEF MSWINDOWS}
  result.AcceptEncoding := 'gzip deflate';
  if result.AcceptEncoding.Contains('gzip') then
    result.AutomaticDecompression := [THTTPCompressionMethod.GZip];
  if result.AcceptEncoding.Contains('deflate') then
    result.AutomaticDecompression := result.AutomaticDecompression + [THTTPCompressionMethod.Deflate];
  {$ENDIF}

  result.ConnectionTimeout := 10000;
  result.ResponseTimeout := 10000;
  result.SendTimeout := 10000;

end;

// URL f[^擾
function GetStreamFromSite(URL: string): TMemoryStream;
var
  HTTPClient: THTTPClient;
begin
  result := TMemoryStream.Create;
  HTTPClient := CreateHttpClient;
  try
    try
      HTTPClient.Get(URL, result);
      if result.Size > 0 then
        result.Seek(0, TSeekOrigin.soBeginning);
    except
      on e: exception do
        FreeAndNil(result);
    end;
  finally
    HTTPClient.Free;
  end;
end;

// URL UTF-8 eLXg擾
function GetTextFromSite(URL: string): string;
var
  Stream: TMemoryStream;
  SList: TStringList;
begin
  result := '';

  Stream := GetStreamFromSite(URL);
  if Stream <> nil then begin
    try
      try
        SList := TStringList.Create;
        try
          SList.LoadFromStream(Stream, TEncoding.UTF8);
          result := SList.Text;
        finally
          SList.Free;
        end;
      except

      end;
   finally
    Stream.Free;
    end;
  end;
end;

// URL摜擾
function GetBitmapFromSite(URL: string): TBitmap;
var
  Stream: TMemoryStream;
begin
  result := nil;

  Stream := GetStreamFromSite(URL);
  if Stream <> nil then begin
    try
      try
        result := TBitmap.Create;
        result.LoadFromStream(Stream);
      except
        on e: exception do begin
          if result <> nil then
            FreeAndNil(result);
        end;
      end;
    finally
      Stream.Free;
    end;
  end;
end;


// ϑ_Xg쐬
function GetSpotList: TList;
var
  Stream: TResourceStream;
  SList: TStringList;
  CommaText: TStringDynArray;
  Text: string;
  Rec, Rec1: PKyoshinRec;
  APoint: TPoint;
  i: integer;
begin
  result := nil;

  if (SpotList <> nil) and (SpotList.Count > 0) then begin
    result := TList.Create;
    for i := 0 to SpotList.Count - 1 do begin
      Rec1 := SpotList.Items[i];
      New(Rec);
      Rec^ := Rec1^;
      result.Add(Rec);
    end;
  end else begin
    Stream := TResourceStream.Create(hInstance, 'SpotList', RT_RCDATA);
    try
      SList := TStringList.Create;
      try
        try
          result := TList.Create;
          SList.LoadFromStream(Stream, TEncoding.UTF8);
          if SList.Count > 2 then begin
            for i := 2 to SList.Count - 1 do begin
              Text := Trim(SList.Strings[i]);
              if not Text.IsEmpty then begin
                CommaText := SplitString(Text, ',');
                if (Length(CommaText) = 19) and Text.StartsWith('1,0') then begin
                  New(Rec);
                  Rec.Index := result.Count;
                  Rec.Name := CommaText[4];
                  Rec.AreaIndex := StrToInt(CommaText[7]);
                  Rec.GeoPoint.X := StrToFloat(CommaText[12]);
                  Rec.GeoPoint.Y := StrToFloat(CommaText[11]);
                  APoint.X := StrToInt(CommaText[15]);
                  APoint.Y := StrToInt(CommaText[16]);
                  Rec.ScanPoint.X := APoint.X +
                                     StrToInt(CommaText[17]);
                  Rec.ScanPoint.Y := APoint.Y +
                                     StrToInt(CommaText[18]);
                  APoint := Get_Ex_MapPos(TPointF.Create(Rec.GeoPoint.X, Rec.GeoPoint.Y));
                  Rec.DispPoint.X := APoint.X;
                  Rec.DispPoint.Y := APoint.Y;
                  Rec.Color := 0;
                  Rec.Scale := $FF;
                  Rec.Distance := 0;
                  result.Add(Rec);
                end;
              end;
            end;
          end;
        except
          on e: exception do
            AddToLog('GetSpotList', e.Message);
        end;
      finally
        SList.Free;
      end;
    finally
      Stream.Free;
    end;
  end;
end;

// ϑ_Xg̃
procedure ClearSpotList(List: TList);
var
  Rec: PKyoshinRec;
begin
  try
    if Assigned(List) then begin
      if List.Count > 0 then
        for Rec in List do
          Dispose(Rec);
      List.Free;
    end;
  except
    on e: exception do
      AddToLog('ClearSpotList', e.Message);
  end;
end;

// GA Xg쐬
// u݂̐kxvŃGAʂɕ\邽

function GetAreaList: TList;
var
  Stream: TResourceStream;
  SList: TStringList;
  SArray: TStringDynArray;
  Rec: PAreaRec;
begin
  result := nil;

  Stream := TResourceStream.Create(hInstance, 'AreaList', RT_RCDATA);
  try
    SList := TStringList.Create;
    try
      try
        SList.LoadFromStream(Stream, TEncoding.UTF8);
        if SList.Count > 0 then begin
          result := TList.Create;
          for var i := 0 to SList.Count - 1 do begin
            SArray := SplitString(Trim(SList.Strings[i]), ',');
            if Length(SArray) = 4 then begin
              New(Rec);
              Rec.Create(i);
              Rec.Code := TryStrToInteger(SArray[0]);
              Rec.Name := SArray[1];
              Rec.APoint := TPointF.Create(TryStrToSingle(SArray[3]), TryStrToSingle(SArray[2]));
              result.Add(Rec);
            end;
          end;
        end;
      except
        on e: exception do
          AddToLog('GetAreaList', e.Message);
      end;
    finally
      SList.Free;
    end;
  finally
    Stream.Free;
  end;
end;

procedure ClearAreaList(List: TList);
var
  Rec: PAreaRec;
begin
  if List <> nil then begin
    try
      if List.Count > 0 then
        for var i := 0 to List.Count - 1 do begin
          Rec := List.Items[i];
          if Rec.NameList <> nil then
            Rec.NameList.Free;
          Dispose(Rec);
        end;
      List.Free;
    except
      on e: exception do
        AddToLog('ClearAreaList', e.Message);
    end;
  end;
end;

procedure ResetAreaList(List: TList);
begin
  if (List <> nil) and (List.Count > 0) then
    try
      for var i := 0 to List.Count - 1 do
        PAreaRec(List.Items[i]).Reset;
      // \[g
      List.SortList(function (Item1, Item2: Pointer): Integer
      begin
        result := CompareValue(PAreaRec(Item1).Index, PAreaRec(Item2).Index);
      end);
    except
      on e: exception do
        AddToLog('ResetAreaList', e.Message);
    end;
end;

procedure SortAreaList(List: TList);
begin
  if (List <> nil) and (List.Count > 0) then
    try
      for var i := 0 to List.Count - 1 do
        PAreaRec(List.Items[i]).SortNameListByScale;
      // ~\[g
      List.SortList(function (Item1, Item2: Pointer): Integer
      begin
        result := CompareValue(PAreaRec(Item2).Scale, PAreaRec(Item1).Scale);
      end);
    except
      on e: exception do
        AddToLog('ResetAreaList', e.Message);
    end;

end;


{  TAreaRec }

constructor TAreaRec.Create(AIndex: integer);
begin
  Self.Index := AIndex;
  Self.Code := 0;
  Self.Name := '';
  Self.APoint := TPointF.Zero;
  Self.Scale := 0;
  Self.Color := 0;
  Self.NameList := nil;
end;

procedure TAreaRec.Free;
begin
  Self.NameList.Free;
end;

procedure TAreaRec.Reset;
begin
  Self.Scale := 0;
  Self.Color := 0;
  FreeAndNil(Self.NameList);
end;

// NameList ̃XP[lɂ~\[g
procedure TAreaRec.SortNameListByScale;
begin
  if (Self.NameList <> nil) and (Self.NameList.Count > 1) then
    Self.NameList.Sort(TComparer<TNameListRec>.Construct(
      function(const Left, Right: TNameListRec): Integer
      begin
        result := CompareValue(Left.Scale, Right.Scale);
      end));
end;


{
  TAverages = record
  private
    Average: array[0..3] of TAverage;
    CurrentPoint: integer;
    CurrentTime: TDateTime;
  public
    procedure Init;
    procedure DoShift;
    function CalcAverage: integer;
    procedure Add(Value: integer; ATime: TDateTime);
  end;
}

function RecodeSecondAndMillisecond(DateTime: TDateTime): TDateTime;
begin
  //result := RecodeMillisecond(DateTime, 0);
  //result := RecodeSecond(result, 0);
  if not TryRecodeDateTime(DateTime,
                           RecodeLeaveFieldAsIs, RecodeLeaveFieldAsIs, RecodeLeaveFieldAsIs,
                           RecodeLeaveFieldAsIs, RecodeLeaveFieldAsIs, 0, 0, result) then
    result := 0;

end;

constructor TAverages.Create(DateTime: TDateTime);
begin
  for var i := 0 to High(Average) do begin
    Average[i].DateTime := 0;
    Average[i].Count := 0;
    Average[i].Sum := 0;
  end;
  CurrentIndex := 0;
  CurrentTime := RecodeSecondAndMillisecond(DateTime);
  Average[CurrentIndex].DateTime := CurrentTime;
end;

procedure TAverages.DoShift(DateTime: TDateTime);
begin
  Inc(CurrentIndex);
  if CurrentIndex > 3 then
    CurrentIndex := 0;

  Average[CurrentIndex].DateTime := DateTime;
  Average[CurrentIndex].Count := 0;
  Average[CurrentIndex].Sum := 0;

  CurrentTime := DateTime;
end;

function TAverages.CalcAverage: integer;
var
  ACount: integer;
  ASum: integer;
begin
  result := 0;

  try
    ACount := 0;
    ASum := 0;
    for var i := Low(Average) to High(Average) do begin
      if MinutesBetween(Average[i].DateTime, CurrentTime) < 4 then begin
        ACount := ACount + Average[i].Count;
        ASum := ASum + Average[i].Sum;
      end;
    end;

    if ACount > 0 then
      result := EnsureRange(ASum div ACount, 0, 99);

  except

  end;
end;

procedure TAverages.Add(Value: integer; DateTime: TDateTime);
var
  ADateTime: TDateTime;
begin
  ADateTime := RecodeSecondAndMillisecond(DateTime);
  if ADateTime <> CurrentTime then
    DoShift(ADateTime);

  Inc(Average[CurrentIndex].Count);
  Inc(Average[CurrentIndex].Sum, Value);
end;

// Windows ȊO
{$IFNDEF MSWINDOWS}
// tHg̃ftHgTCYFireMonkeyKlύX
var
  SystemFontServiceHook: TFMXSystemFontServiceHook = nil;

{ TFMXSystemFontServiceHook }

constructor TFMXSystemFontServiceHook.Create;
begin
  inherited;
  if TPlatformServices.Current.SupportsPlatformService(IFMXSystemFontService,
      IInterface(FOrgFMXSystemFontService)) then begin
    TPlatformServices.Current.RemovePlatformService(IFMXSystemFontService);
    TPlatformServices.Current.AddPlatformService(IFMXSystemFontService, Self);
  end;
end;

function TFMXSystemFontServiceHook.GetDefaultFontFamilyName: String;
begin
  result := FOrgFMXSystemFontService.GetDefaultFontFamilyName
end;

function TFMXSystemFontServiceHook.GetDefaultFontSize: Single;
begin
  {$IFDEF ANDROID}
  result := 12 * (MainForm.ClientWidth / 360);
  {$ELSE}
  result := 12;
  {$ENDIF}
end;

class procedure TFMXSystemFontServiceHook.RegistService;
begin
  if (SystemFontServiceHook = nil) then
    SystemFontServiceHook := TFMXSystemFontServiceHook.Create;
end;

initialization
  TFMXSystemFontServiceHook.RegistService;

{$ENDIF}

end.
