unit KyoshinUnit;

interface

uses
  System.SysUtils, System.Types, System.UITypes, System.Classes, System.Variants,
  FMX.Types, FMX.Graphics, System.DateUtils, System.StrUtils, FMX.Utils,
  System.Net.HttpClient, System.Net.HttpClientComponent, System.Net.URLClient,
  System.NetConsts,
  IdSNTP, IdGlobalProtocols,
  System.JSON, System.Rtti, System.Math.Vectors, System.Math, FMX.Surfaces,
  FMX.Platform, FMX.Clipboard, System.UIConsts, System.IOUtils, System.SyncObjs,
  System.Generics.Collections, System.Generics.Defaults, System.Diagnostics,
  CommonUnit;

type
  // NTP
  TNTPClient = class(TTHread)
  strict private
    FNextTime: TDateTime;
    FAdjustedCount: integer;
    FCount: integer;
    FJapanBias: TDateTime;
    FBias: TDateTime;
    FUTCEnabled: boolean;
    FErrorNext: boolean;
    FLastTimeCheckTIme: TDateTime;
    FLastNTPTime: TDateTime;
    procedure SetNextTime;
    procedure DoNow;
  protected
    procedure Execute; override;
  public
    constructor Create;
  end;

  // lbgANZX
  THTTPAccess = class(TThread)
  strict private
    FHTTPClient: THTTPClient;
    function AccessNow(Rec: PRequestRec): boolean;
  protected
    procedure Execute; override;
  public
    constructor Create;
    destructor Destroy; override;

  end;

  // EEWf[^擾
  TEEWAccess = class(TThread)
  strict private
    FLastAccessTime: TDateTIme;
    FResponseTime: TDateTime;
    FLastReportTime: TDateTime;
    //FEEWRec: PEEWRec;
    FReportList: TStringList;
    function IsNewReport(Rec: PEEWRec): boolean;

    procedure GetEEWJson(Rec: PResponseRec);
    function ParseJson(Text: string): PEEWRec;
  protected
    procedure Execute; override;
  public
    constructor Create;
    destructor Destroy; override;
  end;

  // kj^摜擾
  TKyoshinAccess = class(TThread)
  type
    TActiveState = (asNone, asParent, asChild);
    PActiveRec = ^TActiveRec;
    TActiveRec = record
      IsActive: boolean;
      State: TActiveState;
      Parent: TPoint;
      Childred: TList<TPoint>;
    end;
  const
    ScanPoints: array[0..18] of TPoint = (
                  (X:1; Y:0), (X:2; Y:0), (X:3; Y:0), (X:4; Y:0),
      (X:0; Y:1), (X:1; Y:1), (X:2; Y:1), (X:3; Y:1), (X:4; Y:1),
      (X:0; Y:2), (X:1; Y:2), (X:2; Y:2), (X:3; Y:2),
      (X:0; Y:3), (X:1; Y:3), (X:2; Y:3), (X:3; Y:3),
      (X:0; Y:4), (X:1; Y:4));
  strict private
    FLastAccessTime: TDateTIme;
    FKyoshinResponseTime: TDateTime;
    FRealShindoResponseTime: TDateTime;
    FSpotList: TList;
    FAreaList: TList;
    FSpotsArray: array[0..351, 0..399] of PActiveRec;
    function UpdateSpotList(Rec: PResponseRec): boolean;
    procedure DoMainKyoshin(Rec: PResponseRec);
    procedure DoRealShindoKyoshin(Rec: PResponseRec; ListUpdated: boolean);
    procedure ResetSpotsArray;
  protected
    procedure Execute; override;
  public
    constructor Create;
    destructor Destroy; override;
  end;

  // \zkx摜擾
  TEstIntAccess = class(TThread)
  strict private
    FLastAccessTime: TDateTIme;
    FEstShindoResponseTime: TDateTime;
    procedure DoEstShindo(Rec: PResponseRec);
  protected
    procedure Execute; override;
  public
    constructor Create;
    destructor Destroy; override;
  end;

  TPixelArray = array[0..475, 0..556] of Byte;

  // PSg摜擾
  TPSWavesAccess = class(TThread)
  const
     EpicenterMark: array[0..3, 0..4] of TPoint = (
       ((X:0; Y:0), (X:1; Y:1), (X:2; Y:2), (X:3; Y:3), (X:4; Y:4)),
       ((X:10; Y:0), (X:9; Y:1), (X:8; Y:2), (X:7; Y:3), (X:6; Y:4)),
       ((X:0; Y:10), (X:1; Y:9), (X:2; Y:8), (X:3; Y:7), (X:4; Y:6)),
       ((X:10; Y:10), (X:9; Y:9), (X:8; Y:8), (X:7; Y:7), (X:6; Y:6)));
  strict private
    FLastAccessTime: TDateTIme;
    FPSWavesResponseTime: TDateTime;
    FLastReportNo: integer;
    FLastEpicenter: TPointF;
    FPixelArray_R: array[0..475, 0..556] of boolean;
    FPixelArray_B: array[0..475, 0..556] of boolean;
    FLastPWaveRadius: integer;
    FLastSWaveRadius: integer;

    FOrgBitmap: TBitmap;
    // temp
    WavesList: TStringList;

    procedure DoPSWaves(Rec: PResponseRec);
    function BitmapToByteArray(Rec: PResponseRec): boolean;
    //procedure CopyPixels(ABitmap: TBitmap);
    procedure ScanSPE(Rec: PPsWavesInfo); //var P, S, E: TPoint);
    function CalcPSWaves(Rec: PPsWavesInfo): boolean;
    procedure IfDebug(Rec: PResponseRec; WRec: PPsWavesInfo);
  protected
    procedure Execute; override;
  public
    constructor Create;
    destructor Destroy; override;
  end;



var
  ForceNTP: boolean = false;
  NTPTimeDiffer: TDateTime = 0;
  NTPAdjusted: boolean = false;
  NTPClient: TNTPClient;

  HttpAccesses: array[0..7] of THTTPAccess;
  RequestQueue: TThreadList = nil;
  ResponceQueues: array[TAccessKind] of TThreadList;

  KyoshinIndex: TKyoshinIndex = TKyoshinIndex.KIacmap_s;

  NextAccessTime: TDateTime;   // ɃTCgɃANZX鎞
  NextKyoshinTime: TDateTime;  // 擾摜̃^CX^v

  NoiseThreshold: integer = 8;
  DetectMicroEq: boolean = false;
  MicroEqThreshold: integer = 8;

  // \pJE^
  AccessSuccess: Int64 = 0;
  AccessError: Int64 = 0;
  AccessRetry: Int64 = 0;

  RequestCount: Int64 = 0;

  AcquisitImage: Int64 = 0;
  AcquisitEEW: Int64 = 0;
  AcquisitEst: Int64 = 0;
  AcquisitPS: Int64 = 0;

  SpotsCount: Int64 = 0;
  ColorSearch2Count: Int64 = 0;


implementation

uses
  MainUnit;


{ TNTPClient }

{$IFDEF DEBUG}
const
  NTPLogName = 'NTP.log';
{$ENDIF}

constructor TNTPClient.Create;
begin
  inherited Create(true);
  FreeOnTerminate := false;

  FNextTime := IncSecond(Now, 5);  // N 5bɎ擾
  FAdjustedCount := 0;
  FCount := 0;
  FJapanBias := EncodeDateTime(1899, 12, 30, 9, 0, 0, 0);
  FJapanBias := - FJapanBias;
  FBias := TimeZoneBias;
  FLastNTPTime := 0;

  FUTCEnabled := false;

  AddToLog(TLogType.ltNTP, 'NTPClient start', '');
end;

procedure TNTPClient.Execute;
begin
  FLastTimeCheckTIme := Now;

  while not Terminated and not AppClosing do begin
    try
      // 15bȓ̍Ď擾͔F߂Ȃ
      if (SecondsBetween(FLastNTPTime, Now) > 15) and ((ForceNTP or (FNextTime <= now))) then begin
        DoNow;
        FLastNTPTime := Now;
        SetNextTime;
        ForceNTP := false;
      // VXestep[hŕύXꂽꍇɑΉ
      end else if MilliSecondsBetween(FLastTimeCheckTIme, Now) > 500 then
        ForceNTP := true;

      FLastTimeCheckTIme := Now;

      if Terminated or AppClosing then Exit;

      Sleep(100);
    except
      on e: exception do
        AddToLog('TNTPClient.Execute', e.Message);
    end;
  end;

end;

// ̒Ԃ̃Zbg
procedure TNTPClient.SetNextTime;
const
  NTP_Interval: array[0..3] of integer = (30, 180, 600, 1800);
begin
  if FErrorNext then begin
    FErrorNext := false;
    FNextTime := IncSecond(Now, 180);
    FCount := 1;
  end else if InRange(FCount, 0, 3) then
    FNextTime := IncSecond(Now, NTP_Interval[FCount])
  else
    FNextTime := IncSecond(Now, NTP_Interval[3]);

  Inc(FCount);
end;

// NTPNCAg{
procedure TNTPClient.DoNow;
var
  IdSNTP: TIdSNTP;
  DT: TDateTime;
begin
  IdSNTP := TIdSNTP.Create;
  try
    try
      IdSNTP.Port := 123;
      IdSNTP.Host := 'ntp.jst.mfeed.ad.jp';
      IdSNTP.ReceiveTimeout := 10000;

      IdSNTP.Active := true;
                                                // NTP
      DT := IdSNTP.DateTime;
      NTPTimeDiffer := IdSNTP.AdjustmentTime;   // NTPƃVXeƂ̍
      {$IFDEF MACOS}
      // macos  TIdSNTP oO?
      if NTPTimeDiffer < - EncodeTime(16, 0, 0, 0) then
        NTPTimeDiffer := NTPTimeDiffer + FBias + FBias;
      {$ENDIF}

      if Terminated or AppClosing then exit;

      NTPAdjusted := true;
      IdSNTP.Active := false;

      // UTC time
      if FUTCEnabled then
        NTPTimeDiffer := NTPTimeDiffer + FBias - FJapanBias;

      NTPAdjusted := true;
      Inc(FAdjustedCount);

      AddToLog(TLogType.ltNTP, 'TNTPClient.DoNow', 'adjusted clock');

      Synchronize(procedure
      begin
        MainForm.Edit10.Text := FormatDateTIme('hh:nn:ss', Clock);
        if DT >= Now then
          MainForm.Edit11.Text := FormatDateTIme('ss.zzz', NTPTimeDiffer) + ' sec'
        else
          MainForm.Edit11.Text := FormatDateTIme('-ss.zzz', NTPTimeDiffer) + ' sec';
        MainForm.Edit13.Text := FAdjustedCount.ToString;
      end);

    except
      on e: exception do
        AddToLog(TLogType.ltNetError, 'TNTPClient.DoNow', e.Message);
    end;
  finally
    IdSNTP.Free;
  end;

end;

{ THTTPAccess }

// 8̃CX^XŁAkj^ւ̃ANZXꊇĎ󂯎
// L[ʂăANZXv󂯎AMɊeL[ɌʂԂ
// ̎dg݂𗘗p闝RF
// eNCAg͑O̎Mv̐ۂɊ֌WȂ1bԊuŗv悤ɂ邽
// iʏ̕@ƃ^CAEg܂Œ~Ă܂Ƃj

constructor THTTPAccess.Create;
begin
  inherited Create(true);
  FreeOnTerminate := false;

  FHTTPClient := CreateHttpClient;
  FHTTPClient.ConnectionTimeout := 4000;
  FHTTPClient.ResponseTimeout := 4000;
  FHTTPClient.SendTimeout := 4000;
end;

destructor THTTPAccess.Destroy;
begin
  try
    FHTTPClient.Free;
  except
    on e: exception do
      AddToLog('THTTPAccess.Destroy', e.Message);
  end;

  inherited;
end;

procedure THTTPAccess.Execute;
var
  Rec: PRequestRec;
begin
  while not Terminated and not AppClosing do begin
    try
    // L[mFANGXgΏ
      Rec := PopFromQueue(RequestQueue);
      if Rec <> nil then begin
        try
          // 2bȓ̃NGXĝݏ
          if SecondsBetween(Rec.DateTime, Clock) < 8 then begin
            if Terminated or AppClosing then Exit;
            if not AccessNow(Rec) then begin
              if Terminated or AppClosing then Exit;
              // HTTP֘ÃG[HTTPClinetCX^Đ
              FHTTPClient.Free;
              FHTTPClient := CreateHttpClient;
            end;
          end;
        finally
          Dispose(Rec);
        end;
      end;

      if Terminated or AppClosing then Exit;

      Sleep(100);
    except
      on e: exception do
        AddToLog('THTTPAccess.Execute', e.Message);
    end;
  end;
end;

function THTTPAccess.AccessNow(Rec: PRequestRec): boolean;
var
  Stream: TMemoryStream;
  AHTTPResponse: IHTTPResponse;
  CanGet: boolean;
  ResponseRec: PResponseRec;
  RetryCount: integer;
begin
  result := false;

  if Rec <> nil then begin
   Stream := TMemoryStream.Create;
    try
      CanGet := false;
      RetryCount := 0;
      // HEADŃt@C݂̑mFi100msԊuōő4񎎍sj
      for var i := 0 to 3 do begin
        try
          AHTTPResponse := FHttpClient.Head(Rec.URL);
          if Terminated or AppClosing then exit;
          if AHTTPResponse.StatusCode = 200 then begin
            CanGet := true;
            Break;
          end;
        except

        end;
        Inc(RetryCount);
        if Terminated or AppClosing then exit;
        Sleep(100);
        if Terminated or AppClosing then exit;
      end;

      // gC񐔂L^
      if RetryCount > 0 then
        TInterlocked.Add(AccessRetry, RetryCount);

      // sI
      if not CanGet then begin
        AddToLog(TLogType.ltNetError, 'THTTPAccess.AccessNow', 'Retry Error: ' + AHTTPResponse.StatusText);
        TInterlocked.Increment(AccessError);
        exit;
      end;

      if Terminated or AppClosing then exit;

      try
        AHTTPResponse := FHttpClient.Get(Rec.URL, Stream);
        if Terminated or AppClosing then exit;
        if (AHTTPResponse.StatusCode = 200) and (Stream.Size > 0) then begin
          Stream.Seek(0, TSeekOrigin.soBeginning);
          New(ResponseRec);
          ResponseRec.DateTime := Rec.DateTime;
          ResponseRec.AccessKind := Rec.AccessKind;
          ResponseRec.Stream := Stream;
          ResponseRec.KyoshinIndex := Rec.KyoshinIndex;
          if Terminated or AppClosing then exit;
          // eL[ɎMXg[𑗂
          PushToQueue(ResponceQueues[Rec.AccessKind], ResponseRec);
          result := true;
          // ML^
          TInterlocked.Increment(AccessSuccess);
        end;

      except
        on e: exception do begin
          AddToLog(TLogType.ltNetError, 'THTTPAccess.AccessNow', e.Message);
          // MsL^
          TInterlocked.Increment(AccessError);
        end;
      end;
    finally
      if not result then
        Stream.Free;
    end;
  end;
end;

{ TEEWAccess }

function MakeEEWURL(DateTime: TDateTime): string;
begin
  result := 'http://www.kmoni.bosai.go.jp/webservice/hypo/eew/%DateTime%.json?jQuery(%DateTime%)';
  result := result.Replace('%DateTime%', FormatDateTime('yyyymmddhhnnss', DateTime), [rfReplaceAll]);
end;

constructor TEEWAccess.Create;
begin
  inherited Create(true);
  FreeOnTerminate := false;

  FLastAccessTime := 0;
  FResponseTime := 0;
  FLastReportTime := 0;

  FReportList := TStringList.Create;

end;

destructor TEEWAccess.Destroy;
begin
  try
    FReportList.Free;

  except
    on e: exception do
      AddToLog('TEEWAccess.Destroy', e.Message);
  end;

  inherited;
end;

procedure TEEWAccess.Execute;
var
  RequestRec: PRequestRec;
  ResponseRec: PResponseRec;
begin
  while not Terminated and not AppClosing do begin
    try
      // L[HTTPX|XΏ
      ResponseRec := PopFromQueue(ResponceQueues[TAccessKind.akEEW]);
      if ResponseRec <> nil then begin
        // ŌɎMEukj^ԁvV̂łȂ΂ȂȂ
        if FResponseTime < ResponseRec.DateTime then begin
          FResponseTime := ResponseRec.DateTime;
          GetEEWJson(ResponseRec);
        end;
        if Assigned(ResponseRec.Stream) then
          ResponseRec.Stream.Free;
        Dispose(ResponseRec);
      // ukj^ԁvXVĂAANZXNGXg
      end else if (FLastAccessTime <> NextAccessTime) and (NextAccessTime <= Clock) then begin
        FLastAccessTime := NextAccessTime;
        New(RequestRec);
        RequestRec.DateTime := NextKyoshinTime;
        RequestRec.URL := MakeEEWURL(NextKyoshinTime);
        RequestRec.AccessKind := TAccessKind.akEEW;
        PushToQueue(RequestQueue, RequestRec);
      end;

      if Terminated or AppClosing then exit;

      Sleep(100);
    except
      on e: exception do
        AddToLog('TEEWAccess.Execute', e.Message);
    end;
  end;
end;

// V
function TEEWAccess.IsNewReport(Rec: PEEWRec): boolean;
  procedure AddToReportList;
  begin
    FReportList.Add(Rec.ReportID + '=' + Rec.ReportNo.ToString);
    result := true;
  end;
  procedure UpdateReportList;
  begin
    FReportList.Values[Rec.ReportID] := Rec.ReportNo.ToString;
    result := true;
  end;
var
  ReportNo: string;
begin
  result := false;

  try
    if FReportList.Count > 0 then begin
      ReportNo := FReportList.Values[Rec.ReportID];
      if not ReportNo.IsEmpty then begin
        if TryStrToInteger(ReportNo) < Rec.ReportNo then
          UpdateReportList;
      end else
        AddToReportList;
    end else
      AddToReportList;
  except
    on e: exception do
      AddToLog('TEEWAccess.IsNewReport', e.Message);
  end;
end;

procedure TEEWAccess.GetEEWJson(Rec: PResponseRec);
var
  SList: TstringList;
  Json: string;
  EEWRec: PEEWRec;
begin
  Json := '';

  try
    if Assigned(Rec.Stream) then begin
      SList := TstringList.Create;
      try
        SList.LoadFromStream(Rec.Stream, TEncoding.UTF8);
        Json := Trim(SList.Text);
      finally
        SList.Free;
      end;
    end;

    if Terminated or AppClosing then Exit;

    if not Json.IsEmpty and not Json.Contains('f[^܂') then begin
      // Json p[X
      EEWRec := ParseJson(Json);
      if EEWRec <> nil then begin
        EEWRec.KyoshinTime := FResponseTime;
        // O[oϐXV
        EEW_Rec := EEWRec^;
        if IsNewReport(EEWRec) and not Terminated and not AppClosing  then begin
          Synchronize(procedure
            begin
              // UI XbhŃ|[gXV
              MainForm.UpdateEEWDetails(EEWRec);
            end);
          EEWState := true;
          FLastReportTime := EEWRec.ReportTime;
        end;
        Dispose(EEWRec);
      end;
    end else begin
      // EEW[ĥƂAŌEEW|[gM30bEEW[hI
      if EEWState and (SecondsBetween(FLastReportTime, Clock) > 30) then begin
        EEWState := false;
        FLastReportTime := 0;
        Synchronize(procedure
        begin
          // UI XbhŏI
          MainForm.ResetEEWDetails;
        end);
      end;
    end;
  except
    on e: exception do
      AddToLog('TEEWAccess.GetEEWJson', e.Message);
  end;
end;

// Json p[X
function TEEWAccess.ParseJson(Text: string): PEEWRec;
var
  S: string;
  JSONObject: TJSONObject;
  JSONPair: TJsonValue;
  DepthInt: integer;
begin
  result := nil;

  S := Trim(Text);
  JSONObject := TJSONObject.ParseJSONValue(S) as TJSONObject;
  try
    try
      if JSONObject <> nil then begin
        New(result);
        result^ := System.Default(TEEWRec); // R[h̏
        if JSONObject.TryGetValue('report_time', JsonPair) then
          result.ReportTime := StrToDateTimeEx2(JsonPair.Value);
        if JSONObject.TryGetValue('report_id', JsonPair) then
          result.ReportID := JsonPair.Value;
        if JSONObject.TryGetValue('report_num', JsonPair) then
          result.ReportNo := TryStrToInteger(JsonPair.Value);
        if JSONObject.TryGetValue('is_cancel', JsonPair) then
          result.IsCancel := JsonPair.Value = 'true';
        if JSONObject.TryGetValue('alertflg', JsonPair) then
          result.IsAlert := JsonPair.Value = 'true';
        if JSONObject.TryGetValue('is_final', JsonPair) then
          result.IsFinal := JsonPair.Value = 'true';
        if JSONObject.TryGetValue('is_training', JsonPair) then
          result.IsTraining := JsonPair.Value = 'true';
        if JSONObject.TryGetValue('origin_time', JsonPair) then
          result.OriginTIme := StrToDateTimeEx(JsonPair.Value);
        if JSONObject.TryGetValue('region_code', JsonPair) then
          result.EpicenterCode := TryStrToInteger(JsonPair.Value);
        if JSONObject.TryGetValue('region_name', JsonPair) then
          result.EpicenterName := JsonPair.Value;

        if JSONObject.TryGetValue('latitude', JsonPair) then
          result.EpicenterF.Y := TryStrToSingle(JsonPair.Value);
        if JSONObject.TryGetValue('longitude', JsonPair) then
          result.EpicenterF.X := TryStrToSingle(JsonPair.Value);

        result.Epicenter := Get_Ex_MapPos(result.EpicenterF);

        if JSONObject.TryGetValue('depth', JsonPair) then
          result.Depth := JsonPair.Value;
        if not result.Depth.IsEmpty then begin
          if not TryStrToInt(Trim(ReplaceText(result^.Depth, 'km', '')), DepthInt) then
            result^.DepthInt := 0
          else
            result^.DepthInt := DepthInt;
        end;

        if JSONObject.TryGetValue('magunitude', JsonPair) then
          result.Magnitude := TryStrToSingle(JsonPair.Value);
        if JSONObject.TryGetValue('calcintensity', JsonPair) then
          result.MaxShindo := JsonPair.Value;

        // 擾L^
        TInterlocked.Increment(AcquisitEEW);
      end;
    except
      on e: exception do
        AddToLog('TEEWAccess.ParseJson', e.Message);
    end;
  finally
    JSONObject.Free;
  end;
end;


{ TKyoshinAccess }

constructor TKyoshinAccess.Create;
begin
  inherited Create(true);
  FreeOnTerminate := false;

  FLastAccessTime := 0;
  FKyoshinResponseTime := 0;
  FRealShindoResponseTime := 0;

  FSpotList := GetSpotList;
  var Rec: PKyoshinRec;
  for Rec in FSpotList do
    Rec.Average.Create(Clock);

  FAreaList := GetAreaList;

  for var y := 0 to 399 do
    for var x := 0 to 351 do begin
      New(FSpotsArray[x, y]);
      FSpotsArray[x, y].Childred := nil;
    end;

end;

destructor TKyoshinAccess.Destroy;
begin
  try
    ClearSpotList(FSpotList);
    ClearAreaList(FAreaList);

    for var y := 0 to 399 do
      for var x := 0 to 351 do
        Dispose(FSpotsArray[x, y]);
  except
    on e: exception do
      AddToLog('TKyoshinAccess.Destroy', e.Message);
  end;

  inherited;
end;

function MakeKyoshinURL(DateTime: TDateTIme; KIndex: TKyoshinIndex): string;
begin
  result := 'http://www.kmoni.bosai.go.jp/data/map_img/RealTimeImg/%KyoshinIndex%/%Date%/%DateTime%.%KyoshinIndex%.gif';
  result := result.Replace('%Date%', FormatDateTime('yyyymmdd', DateTime));
  result := result.Replace('%DateTime%', FormatDateTime('yyyymmddhhnnss', DateTime));
  result := result.Replace('%KyoshinIndex%', KyoshinIndexURLStr[KIndex], [rfReplaceAll]);
end;

procedure TKyoshinAccess.Execute;
var
  RequestRec: PRequestRec;
  ResponseRec: PResponseRec;
begin
  while not Terminated and not AppClosing do begin
    try
      // L[HTTPX|XΏ
      // ŌɎMEukj^ԁvV̂łȂ΂ȂȂ
      // T[ȍԂɂAŌɎMEukj^ԁvÂ̂ԂĂ邱Ƃ
      ResponseRec := PopFromQueue(ResponceQueues[TAccessKind.akKyoshin]);
      if ResponseRec <> nil then begin
        if ResponseRec.AccessKind = TAccessKind.akKyoshin then begin
          // ʏ̋kj^
          if ResponseRec.KyoshinIndex = KyoshinIndex then begin
            if FKyoshinResponseTime < ResponseRec.DateTime then begin
              FKyoshinResponseTime := ResponseRec.DateTime;
              DoMainKyoshin(ResponseRec);
            end;
          // A^Ckx
          end else if ResponseRec.KyoshinIndex = TKyoshinIndex.KIjma_s then begin
            if FRealShindoResponseTime < ResponseRec.DateTime then begin
              FRealShindoResponseTime := ResponseRec.DateTime;
              DoRealShindoKyoshin(ResponseRec, false);
            end;
          end;
        end;
        // uṽ
        if Assigned(ResponseRec.Stream) then
          ResponseRec.Stream.Free;
        Dispose(ResponseRec);

        // 擾L^
        TInterlocked.Increment(AcquisitImage);

      // ukj^ԁvXVĂAANZXNGXg
      end else if (FLastAccessTime <> NextAccessTime) and (NextAccessTime <= Clock) then begin
        FLastAccessTime := NextAccessTime;
        New(RequestRec);
        RequestRec.DateTime := NextKyoshinTime;
        RequestRec.URL := MakeKyoshinURL(NextKyoshinTime, KyoshinIndex);
        RequestRec.AccessKind := TAccessKind.akKyoshin;
        RequestRec.KyoshinIndex := KyoshinIndex;
        PushToQueue(RequestQueue, RequestRec);
        // KyoshinIndex n\kxłȂꍇAn\kx摜ǉŎ擾
        if KyoshinIndex <> TKyoshinIndex.KIjma_s then begin
          New(RequestRec);
          RequestRec.DateTime := NextKyoshinTime;
          RequestRec.URL := MakeKyoshinURL(NextKyoshinTime, TKyoshinIndex.KIjma_s);
          RequestRec.AccessKind := TAccessKind.akKyoshin;
          RequestRec.KyoshinIndex := TKyoshinIndex.KIjma_s;
          PushToQueue(RequestQueue, RequestRec);
        end;
      end;

      if Terminated or AppClosing then Exit;

      Sleep(100);
    except
      on e: exception do
        AddToLog('TKyoshinAccess.Execute', e.Message);
    end;
  end;

end;

// Cf[^ eϑ_̐F擾AScalel(0..99)ɊZ
function TKyoshinAccess.UpdateSpotList(Rec: PResponseRec): boolean;
var
  ABitmap: TBitmap;
  BitmapData: TBitmapData;
  KRec: PKyoshinRec;
  Counter: integer;
begin
  result := false;

  Counter := 0;
  if (Rec.Stream <> nil) and (Rec.Stream.size > 1024) then begin
    ABitmap := TBitmap.Create;
    try
      ABitmap.LoadFromStream(Rec.Stream);
      if ABitmap.Width = 352 then begin
        if ABitmap.Map(TMapAccess.Read, BitmapData) then begin
          try
            if FSpotList.Count > 0 then begin
              try
                for var i := 0 to FSpotList.Count - 1 do begin
                  if Terminated or AppCLosing then Exit;
                  KRec := FSpotList.Items[i];
                  KRec.Color := BitmapData.GetPixel(KRec.ScanPoint.X, KRec.ScanPoint.Y);
                  KRec.AveragedScale := 0;
                  if IsGray(KRec.Color) then begin
                    KRec.Scale := -1;
                  end else begin
                    KRec.Scale := ColorToIndexScale(KRec.Color);
                    if Rec.KyoshinIndex = TKyoshinIndex.KIacmap_s then begin
                      Inc(Counter);
                      // Average  add 2023-07-23
                      if not EEWState then
                        KRec.Average.Add(KRec.Scale, Rec.DateTime);    // EEWɂ͕ϒlύXȂ
                      if KRec.Scale > KRec.Average.CalcAverage + NoiseThreshold then begin
                        KRec.AveragedScale := KRec.Scale;
                      end else
                        KRec.AveragedScale := 0;
                    end;
                  end;
                end;
                result := true;
              except
                on e: exception do
                  AddToLog('TKyoshinAccess.UpdateSpotList', e.Message);
              end;
            end;
          finally
            ABitmap.UnMap(BitmapData);
          end;
        end;
        if Rec.KyoshinIndex = TKyoshinIndex.KIacmap_s then
          TInterlocked.Exchange(SpotsCount, Counter);
      end;
    finally
      ABitmap.Free;
    end;
  end;
end;

procedure TKyoshinAccess.DoMainKyoshin(Rec: PResponseRec);
var
  AStopWatch: TStopWatch;
begin
  if Rec.KyoshinIndex = TKyoshinIndex.KIacmap_s then
    AStopWatch := TStopWatch.StartNew;

  if UpdateSpotList(Rec) then begin
    try
      // \[g
      FSpotList.SortList(function (Item1, Item2: Pointer): Integer
        begin
          result := CompareValue(PKyoshinRec(Item1).Scale, PKyoshinRec(Item2).Scale);
        end);

      if Terminated or AppClosing then exit;

      // UI XV
      Synchronize(procedure
      begin
        MainForm.UpdateKyoshinSpots(FSpotList, Rec.DateTime, Rec.KyoshinIndex);
        // FTъϑ_`̏ԁilbgAbNZXԂj
        if Rec.KyoshinIndex = TKyoshinIndex.KIacmap_s then
          MainForm.Edit32.Text := AStopWatch.ElapsedMilliseconds.ToString;
      end);

      if Terminated or AppClosing then exit;

      // A^Ckx̂Ƃ
      if Rec.KyoshinIndex = TKyoshinIndex.KIjma_s then
        DoRealShindoKyoshin(Rec, true);

    except
      on e: exception do
        AddToLog('TKyoshinAccess.DoMainKyoshin', e.Message);
    end;
  end;
end;

// A^Ckx̏
procedure TKyoshinAccess.DoRealShindoKyoshin(Rec: PResponseRec; ListUpdated: boolean);
var
  KRec: PKyoshinRec;
  ARec: PAreaRec;
  NRec: TNameListRec;
  AreaListCount: integer;
  AScale: integer;
  AColor: TAlphaColor;
begin
  if not ListUpdated then begin
    if not UpdateSpotList(Rec) then exit
  end;

  try
    // GAƂɏWv
    ResetAreaList(FAreaList);
    for var i := 0 to FSpotList.count - 1 do begin
      KRec := FSpotList.Items[i];
      if KRec.Scale >= MinShindoScale then begin  // ݒŏkxƔrAȏł
        ARec := FAreaList.Items[KRec.AreaIndex];
        if ARec.Scale < KRec.Scale then begin
          ARec.Scale := KRec.Scale;               // ōl̕ۑ
          ARec.Color := KRec.Color;
          if ARec.NameList = nil then             // ϑ_L^
            ARec.NameList := TList<TNameListRec>.Create;
          NRec.Name := KRec.Name;
          NRec.Scale := KRec.Scale;
          ARec.NameList.Add(NRec);
        end;
      end;
    end;

    if Terminated or AppClosing then exit;

    // \GAvZ
    SortAreaList(FAreaList);                      // GAXg̍~\[g
    AreaListCount := 0;
    for var i := 0 to 3 do begin                  // ő4GA܂
      ARec := FAreaList.Items[i];
      if ARec.Scale >= MinShindoScale then        // ݒŏkxȏł
        Inc(AreaListCount);
    end;

    if Terminated or AppClosing then exit;

    // SpotListIndex\[g
    FSpotList.SortList(function (Item1, Item2: Pointer): Integer
      begin
        result := CompareValue(PKyoshinRec(Item1).Index, PKyoshinRec(Item2).Index);
      end);

    if Terminated or AppClosing then exit;

    // ݒn̐kx
    AScale := -1;
    AColor := $FFE0E0E0;
    for var i := 0 to 3 do begin                  // ݒn̍ŋߊϑ_̒lZbg
      KRec := FSpotList[NearestSpots[i].Index];
      if KRec.Scale > 0 then begin
        AScale := KRec.Scale;
        AColor := KRec.Color;
        Break;
      end;
    end;

    if Terminated or AppClosing then exit;

    // UI XV
    Synchronize(procedure
    begin
      MainForm.UpdateShindoList(FAreaList, AreaListCount, AScale, AColor);
    end);

  except
    on e: exception do
      AddToLog('TKyoshinAccess.DoRealShindoKyoshin', e.Message);
  end;
end;

procedure TKyoshinAccess.ResetSpotsArray;
begin
  try
    for var y := 0 to 399 do begin
      for var x := 0 to 351 do begin
        FSpotsArray[x, y].IsActive := false;
        FSpotsArray[x, y].State := asNone;
        FreeAndNIl(FSpotsArray[x, y].Childred);
      end;
    end;
  except
    on e: exception do
      AddToLog('TKyoshinAccess.ResetSpotsArray', e.Message);
  end;

end;

{  TEstIntAccess }

constructor TEstIntAccess.Create;
begin
  inherited Create(true);
  FreeOnTerminate := false;

  FLastAccessTime := 0;
  FEstShindoResponseTime := 0;

end;

destructor TEstIntAccess.Destroy;
begin

  inherited;
end;

function MakeEstShindoImgURL(DateTime: TDateTime): string;
begin
  result := 'http://www.kmoni.bosai.go.jp/data/map_img/EstShindoImg/eew/%Date%/%DateTime%.eew.gif';
  result := result.Replace('%Date%', FormatDateTime('yyyymmdd', DateTime));
  result := result.Replace('%DateTime%', FormatDateTime('yyyymmddhhnnss', DateTime));
end;

procedure TEstIntAccess.Execute;
var
  RequestRec: PRequestRec;
  ResponseRec: PResponseRec;
begin
  while not Terminated and not AppClosing do begin
    try
      // L[HTTPX|XΏ
        // ŌɎMEukj^ԁvV̂łȂ΂ȂȂ
        // T[ȍԂɂAŌɎMEukj^ԁvÂ̂ԂĂ邱Ƃ
      // \zkx摜̉
      ResponseRec := PopFromQueue(ResponceQueues[TAccessKind.akEstInt]);
      if ResponseRec <> nil then begin
        if EEWState and (FEstShindoResponseTime < ResponseRec.DateTime) then begin
              FEstShindoResponseTime := ResponseRec.DateTime;
              DoEstShindo(ResponseRec);
        end;
        // uṽ
        if Assigned(ResponseRec.Stream) then
          ResponseRec.Stream.Free;
        Dispose(ResponseRec);

        // 擾L^
        TInterlocked.Increment(AcquisitEst);

      // ukj^ԁvXVĂAANZXNGXg
      end else if EEWState and (FLastAccessTime <> NextAccessTime) and (NextAccessTime <= Clock) then begin
        FLastAccessTime := NextAccessTime;
        // \zkx摜̗v
        New(RequestRec);
        RequestRec.DateTime := NextKyoshinTime;
        RequestRec.URL := MakeEstShindoImgURL(NextKyoshinTime);
        RequestRec.AccessKind := TAccessKind.akEstInt;
        // RequestRec.KyoshinIndex := KyoshinIndex;
        PushToQueue(RequestQueue, RequestRec);
      end;

      if Terminated or AppClosing then exit;

      Sleep(100);
    except
      on e: exception do
        AddToLog('TEstIntAccess.Execute', e.Message);
    end;
  end;
end;

procedure TEstIntAccess.DoEstShindo(Rec: PResponseRec);
var
  ABitmap: TBitmap;
  Scale: integer;
  BitmapData: TBitmapData;
  AColor: TAlphaColor;
  NRec: TNearestSpot;
begin
  Scale := -1;
  AColor := TAlphaColorRec.Null;

  ABitmap := TBitmap.Create(352, 400);
  try
    if Rec.Stream <> nil then begin
      try
        ABitmap.LoadFromStream(Rec.Stream);
        if ABitmap.Width = 352 then begin
          if ABitmap.Map(TMapAccess.Read, BitmapData) then begin
            try
              // ݒn̈ʒuɂF擾
              AColor := BitmapData.GetPixel(CurrentPoint.X, CurrentPoint.Y);
              // FȂꍇ͍ŋߊϑ_擾 ... KvH
              if AColor = 0 then begin
                for var i := 0 to 3 do begin
                  NRec := NearestSpots[i];
                  AColor := BitmapData.GetPixel(NRec.ScanPoint.X, NRec.ScanPoint.Y);
                  if AColor <> 0 then
                    Break;
                end;
              end;

             if Terminated or AppClosing then exit;

              // FXP[l擾
              if not IsGray(AColor) then begin
                Scale := ColorToIndexScale(AColor);
                if not InRange(Scale, 0, 99) then
                  Scale := -1;
              end else
                Scale := -1;
            finally
              ABitmap.UnMap(BitmapData);
            end;
          end;
        end;
      except
        on e: exception do
          AddToLog('TEstIntAccess.DoEstShindo', e.Message);
      end;
    end;
  finally
    ABitmap.Free;
  end;

  if Terminated or AppClosing then exit;

  Synchronize(procedure
  begin
    MainForm.UpdateEstShindo(Scale, AColor);
  end);
end;



{ TPSWavesAccess }

// EEW ̌ݒn̗\zkxPSg̃XLƌvZ
constructor TPSWavesAccess.Create;
begin
  inherited Create(true);
  FreeOnTerminate := false;

  FLastAccessTime := 0;
  FPSWavesResponseTime := 0;
  FLastReportNo := 0;
  FLastEpicenter := TPointF.Zero;

  FLastPWaveRadius := 0;
  FLastSWaveRadius := 0;

  FOrgBitmap := TBitmap.Create;
  WavesList := TStringList.Create;
end;

destructor TPSWavesAccess.Destroy;
begin
  try
    FOrgBitmap.Free;
    WavesList.Free;
  except
    on e: exception do
      AddToLog('TPSWavesAccess.Destroy', e.Message);
  end;

  inherited;
end;

function MakePSWaveImgURL(DateTime: TDateTime): string;
begin
  result := 'http://www.kmoni.bosai.go.jp/data/map_img/PSWaveImg/eew/%Date%/%DateTime%.eew.gif';
  result := result.Replace('%Date%', FormatDateTime('yyyymmdd', DateTime));
  result := result.Replace('%DateTime%', FormatDateTime('yyyymmddhhnnss', DateTime));
end;

procedure TPSWavesAccess.Execute;
var
  RequestRec: PRequestRec;
  ResponseRec: PResponseRec;
begin
  while not Terminated and not AppClosing do begin
    try
      if not EEWState then begin
        FLastReportNo := 0;
        // temp
        if IsDebug then begin
          if WavesList.Count > 0 then begin
            WavesList.SaveToFile(AppDir + PathDelim + FormatDateTime('hhnnss', Clock) + 'PSWaves.txt', TEncoding.UTF8);
            WavesList.Clear;
          end;
        end;
      end;

      // L[HTTPX|XΏ
        // ŌɎMEukj^ԁvV̂łȂ΂ȂȂ
        // T[ȍԂɂAŌɎMEukj^ԁvÂ̂ԂĂ邱Ƃ
      // \zkx摜̉
      ResponseRec := PopFromQueue(ResponceQueues[TAccessKind.akPSWaves]);
      if ResponseRec <> nil then begin
        if EEWState and (FPSWavesResponseTime < ResponseRec.DateTime) then begin
              FPSWavesResponseTime := ResponseRec.DateTime;
              DoPSWaves(ResponseRec);
        end;
        // uṽ
        if Assigned(ResponseRec.Stream) then
          ResponseRec.Stream.Free;
        Dispose(ResponseRec);

        // 擾L^
        TInterlocked.Increment(AcquisitPS);

      // ukj^ԁvXVĂAANZXNGXg
      end else if EEWState and (FLastAccessTime <> NextAccessTime) and (NextAccessTime <= Clock) then begin
        FLastAccessTime := NextAccessTime;
        // PSg摜̗v
        New(RequestRec);
        RequestRec.DateTime := NextKyoshinTime;
        RequestRec.URL := MakePSWaveImgURL(NextKyoshinTime);
        RequestRec.AccessKind := TAccessKind.akPSWaves;
        // RequestRec.KyoshinIndex :=
        PushToQueue(RequestQueue, RequestRec);
      end;

      if Terminated or AppClosing then exit;

      Sleep(100);
    except
      on e: exception do
        AddToLog('TPSWavesAccess.Execute', e.Message);
    end;
  end;

end;

// kj^PSg摜PgSg̐k̋
// SgݒnɓB鎞ԂZo
// alter 2023-07-27
procedure TPSWavesAccess.DoPSWaves(Rec: PResponseRec);
var
  WaitingCount: integer;
  WRec: PPsWavesInfo;
begin
  if Rec.Stream <> nil then begin
      try
        New(WRec);
        try
          WRec^ := System.Default(TPsWavesInfo);   // 

          // kj^PSg摜bytezɕϊ
          if not BitmapToByteArray(Rec) then
            exit;

          // oCgz񂩂kAPgASg̈ʒuXL
          ScanSPE(WRec);

          // XLʂ̑Ó`FbN
          // CheckPSWaves(WRec);   // {̓`FbNsł邪{Avł͏ȗ

          if Terminated or AppClosing then Exit;

          // kXLłĂ
          if not WRec.Epicenter.IsZero then begin
            WRec.EpicenterF := Get_Ex_LonLan(WRec.Epicenter);
            WRec.EpicenterScanned := true;
          end else begin
            // n}OɐkꍇȂǁAk̃XLłȂƂ
            // EEW f[^PSg摜̃^CX^vv܂ōő 0.5sec ҋ@
            // ^CX^vEEWf[^kZbg
            WRec.EpicenterScanned := false;
            WaitingCount := 0;
            while FPSWavesResponseTime > EEW_Rec.KyoshinTime do begin
              Sleep(50);
              if Terminated or AppClosing then Exit;
              Inc(WaitingCount);
              if WaitingCount > 10 then begin
                if IsDebug then
                  AddToLog(TLogType.ltDebug, 'TPSWavesAccess.DoPSWaves', 'Epicenter Error');
                Exit;
              end;
            end;

            WRec.EpicenterF := EEW_Rec.EpicenterF;
            WRec.Epicenter := Get_Ex_MapPos(WRec.EpicenterF);
            WRec.EpicenterScanned := false;
          end;

          if CalcPSWaves(WRec) then begin
            Synchronize(procedure
            begin
              MainForm.UpdatePSWaves(WRec);
            end);

            if Terminated or AppClosing then Exit;

            if IsDebug then
              IfDebug(Rec, WRec);
          end;

        finally
          Dispose(WRec);
        end;
      except
        on e: exception do
          AddToLog('TPSWavesAccess.DoPSWaves', e.Message);
      end;
    end;
end;

procedure TPSWavesAccess.IfDebug(Rec: PResponseRec; WRec: PPsWavesInfo);
{$IFDEF MSWINDOWS}
var
  ABitmap: TBitmap;
  BitmapData: TBitmapData;
  Line: PAlphaColorArray;
      {$ENDIF}
begin
  try
    if not EEW_Rec.ReportID.IsEmpty then begin
      // save source gif
      var Filename := CaptureDir + PathDelim + EEW_Rec.ReportID + 'PS_Src';
      if not TDirectory.Exists(Filename) then
        TDirectory.CreateDirectory(Filename);
      Filename := Filename + PathDelim + FormatDateTime('yyyymmdd-hhnnss', Rec.DateTime) + '.png';
      FOrgBitmap.SaveToFile(Filename);

      {$IFDEF MSWINDOWS}
      if Terminated or AppClosing then Exit;
      // set and save PS waves Bitmap
      Filename := CaptureDir + PathDelim + EEW_Rec.ReportID + 'PS_Ex';
      if not TDirectory.Exists(Filename) then
        TDirectory.CreateDirectory(Filename);
      ABitmap := TBitmap.Create(Ex_MapSize.cx, Ex_MapSize.cy);
      try
        ABitmap.Clear(0);
        if ABitmap.Map(TMapAccess.Write, BitmapData) then
          try
            for var Y := 0 to BitmapData.Height - 1 do begin
              Line := BitmapData.GetScanline(Y);
              for var X := 0 to BitmapData.Width - 1 do begin
                if FPixelArray_R[X, Y] then
                  Line^[X] := TAlphaColorRec.Red
                else if FPixelArray_B[X, Y] then
                  Line^[X] := TAlphaColorRec.Blue;
              end;
            end;
          finally
            ABitmap.UnMap(BitmapData);
          end;
        FileName := FIleName + Pathdelim + FormatDateTime('yyyymmdd-hhnnss', Rec.DateTime) + '.png';
        ABitmap.SaveToFile(Filename)
      finally
        ABitmap.Free;
      end;
      {$ENDIF}

    end;

    if Terminated or AppClosing then Exit;

    // add params for save
    var PDistance, SDistance: single;
    SecondToDistance(SecondsBetween(EEW_Rec.OriginTIme, Rec.DateTime), EEW_Rec.DepthInt, PDistance, SDistance);
    WavesList.Add(FormatDateTime('yyyy-mm-dd hh:nn:ss', Rec.DateTime) + ' ' +
      'PWave=' + WRec.PWave.X.ToString + ',' + WRec.PWave.Y.ToString + ' ' +
      'SWave=' + WRec.SWave.X.ToString + ',' + WRec.SWave.Y.ToString + ' ' +
      'Epicenter=' + WRec.Epicenter.X.ToString + ',' + WRec.Epicenter.Y.ToString + ' ' +
      'EpicenterF=' + WRec.EpicenterF.X.ToString(TFloatFormat.ffFixed, 7, 1) + ',' + WRec.EpicenterF.Y.ToString(TFloatFormat.ffFixed, 7, 1) + ' ' +
      'PRadius=' + Round(WRec.Epicenter.Distance(WRec.PWave)).ToString + ' ' +
      'SRadius=' + Round(WRec.Epicenter.Distance(WRec.SWave)).ToString + ' ' +
      'PCalcRadius=' + WRec.PRadius.ToString + ' ' +
      'SCalcRadius=' + WRec.SRadius.ToString + ' ' +
      'PDistance=' + Round(PDistance).ToString + 'km ' +
      'SDistance=' + Round(SDistance).ToString + 'km ' +
      'EpicenterScanned=' + BooleanStr[WRec.EpicenterScanned]);
  except
    on e: exception do
      AddToLog('TPSWavesAccess.IfDebug', e.Message);
  end;
end;

// new 2023-07-20
function TPSWavesAccess.BitmapToByteArray(Rec: PResponseRec): boolean;
var
  BitmapData: TBitmapData;
  SrcLine: PAlphaColorArray;
  X, Y, XX, YY: integer;
  AColorRec: TAlphaColorRec;
begin
  result := false;

  try
    FOrgBitmap.LoadFromStream(Rec.Stream);
    if FOrgBitmap.Width = MapSize.cx then begin
      // clear
      for y := 0 to 556 do
        for x := 0 to 475 do begin
          FPixelArray_R[X, Y] := false;
          FPixelArray_B[X, Y] := false;
        end;

      // Map to ByteArray
      if FOrgBitmap.Map(TMapAccess.Read, BitmapData) then begin
        try
          for Y := 0 to BitmapData.Height - 1 do begin
            SrcLine := BitmapData.GetScanline(Y);
            for X := 0 to BitmapData.Width - 1 do begin
              if (SrcLine^[X] AND $FFFFFF) > 0 then begin
                {$IFDEF MSWINDOWS}
                AColorRec.Color := SrcLine^[X];
                {$ELSE}
                AColorRec.Color := RGBToBGR(SrcLine^[X]);
                {$ENDIF}
                if MaskArray[X, Y] = Byte(TAreaRange.arFrame) then begin
                  YY := Y + 349;
                  if InRange(YY, 0, 556) then begin
                    FPixelArray_R[X, YY] := AColorRec.R > 0;
                    FPixelArray_B[X, YY] := AColorRec.B > 0;
                  end;
                end else if MaskArray[X, Y] = Byte(TAreaRange.arMain) then begin
                  XX := X + 124;
                  if InRange(XX, 0, 475) then begin
                    FPixelArray_R[XX, Y] := AColorRec.R > 0;
                    FPixelArray_B[XX, Y] := AColorRec.B > 0;
                  end;
                end;
              end;
            end;
          end;
        finally
          FOrgBitmap.UnMap(BitmapData);
        end;
        result := true;
      end;
    end;
  except
    on e: exception do
      AddToLog('BitmapToByteArray', e.Message);
  end;
end;

// 擾f[^PgASgAk̈ʒuXL
procedure TPSWavesAccess.ScanSPE(Rec: PPsWavesInfo);
  // ḱmXn}[NXL(1/2\Ă΃XLł)
  function IsEpicenterMatch(X, Y: integer; var Offset: TPoint): boolean;
  var
    MatchRate: array[0..3] of integer;
    XX, YY: integer;
    MatchSuccess: integer;
  begin
    result := false;

    MatchSuccess := 4;
    for var i := 0 to 3 do begin
      MatchRate[i] := 0;
      for var j := 0 to 4 do begin
        XX := X + EpicenterMark[i, j].X;
        YY := Y + EpicenterMark[i, j].Y;
        // if (XX < Ex_MapSize.cx) and (YY < Ex_MapSize.cy) then begin
        // alter 2023-07-25
        if InRange(XX, 0, Ex_MapSize.cx - 1) and InRange(YY, 0, Ex_MapSize.cy - 1) then begin
          if FPixelArray_R[XX, YY] then
            MatchRate[i] := MatchRate[i] + 1
          else
            Break;
        end else
          Break;
      end;
      // add 2023-07-25
      if MatchRate[i] < 5 then begin
        Dec(MatchSuccess);
        if MatchSuccess < 2 then
          exit;
      end;
    end;

    if (MatchRate[0] = 5) and (MatchRate[1] = 5)  then begin
       Offset.X := 5;
       Offset.Y := 5;
       result := true;
    end else if (MatchRate[2] = 5) and (MatchRate[3] = 5) then begin
       Offset.X := 5;
       Offset.Y := -1;
       result := true;
    end else if (MatchRate[0] = 5) and (MatchRate[2] = 5) then begin
       Offset.X := 5;
       Offset.Y := 5;
       result := true;
    end else if (MatchRate[1] = 5) and (MatchRate[3] = 5) then begin
       Offset.X := -1;
       Offset.Y := 5;
       result := true;
    end;

  end;
var
  X, Y, X0, X1: integer;
  Offset: TPoint;
begin
  try
    // epicenter
    Rec.Epicenter := TPoint.Zero;
    for Y := -6 to Ex_MapSize.cy - 6 do begin
      for X := -6 to Ex_MapSize.cx - 6 do begin
        if IsEpicenterMatch(X, Y, Offset) then begin
          Rec.Epicenter := Point(X + Offset.X, Y + Offset.Y);
          Break;
        end;
      end;
      if not Rec.Epicenter.IsZero then
        Break;
    end;

    // PSǧXL邽߁AkXẂ}13͈̔͂𐂒ɏォ牺܂ŃNA
    if not Rec.Epicenter.IsZero then begin
      X0 := Max(0, Rec.Epicenter.X - 2);
      X1 := Min(Ex_MapSize.cx - 1, Rec.Epicenter.X + 11 + 2);
      for Y := 0 to Ex_MapSize.cy - 1 do begin
        for X := X0 to X1 do begin
          FPixelArray_R[X, Y] := false;
          FPixelArray_B[X, Y] := false;
        end;
      end;
    end;

    // PgXLiPɍォj
    Rec.PWave := TPoint.Zero;
    for Y := 0 to Ex_MapSize.cy - 1 do begin
      for X := 0 to Ex_MapSize.cx - 1 do begin
        if FPixelArray_B[X, Y] then begin
          Rec.PWave := Point(X, Y);
          Break;
        end;
      end;
      if not Rec.PWave.IsZero then
        Break;
    end;

    // SgXL
    Rec.SWave := TPoint.Zero;
    for Y := 0 to Ex_MapSize.cy - 8 do begin
      for X := 0 to Ex_MapSize.cx - 8 do begin
        if FPixelArray_R[X, Y] then begin
          Rec.SWave := Point(X, Y);
          Break;
        end;
      end;
      if not Rec.SWave.IsZero then
        Break;
    end;

  except
    on e: exception do
      AddToLog('TPSWavesAccess.ScanSPE', e.Message);
  end;
end;

// ꂽPgASgAk̈ʒuevZ
function TPSWavesAccess.CalcPSWaves(Rec: PPsWavesInfo): boolean;
begin
  result := false;

  try
    // k
    if Rec.Epicenter.IsZero and Rec.EpicenterF.IsZero then
      exit;

    // k猻ݒn܂ł̋ikmj
    var EpicenterToCurrentPoint := Rec.Epicenter.Distance(CurrentPoint);
    // var EpicenterToCurrentPointF := GetHubenyDistance(Rec.EpicenterF, CurrentPointF);
    Rec.EpicenterToCurrPointF := Round(GetHubenyDistance(Rec.EpicenterF, CurrentPointF));

    // kPgikm / pxj
    if not Rec.PWave.IsZero then begin
      // Rec.PDistance := Round(GetHubenyDistance(Rec.EpicenterF, Get_Ex_LonLan(Rec.PWave)));
      Rec.PRadius := Round(Rec.Epicenter.Distance(Rec.PWave));
      Rec.PDistance := Round(Rec.EpicenterToCurrPointF * (Rec.PRadius / EpicenterToCurrentPoint));
    end else begin
      Rec.PDistance := 0;
      Rec.PRadius := 0;
    end;

    // kSgikm / pxj
    if not Rec.SWave.IsZero then begin
      // Rec.SDistance := Round(GetHubenyDistance(Rec.EpicenterF, Get_Ex_LonLan(Rec.SWave)));
      Rec.SRadius := Round(Rec.Epicenter.Distance(Rec.SWave));
      Rec.SDistance := Round(Rec.EpicenterToCurrPointF * (Rec.SRadius / EpicenterToCurrentPoint));
    end else begin
      Rec.SDistance := 0;
      Rec.SRadius := 0;
    end;

    // ܂ł̑ibj
    if EEW_Rec.OriginTime > 0 then
      Rec.TravelTime := SecondsBetween(EEW_Rec.OriginTime, FPSWavesResponseTime)
    else
      Rec.TravelTime := 0;

    // SǧݒnB܂ł̎cb   ̏̓O[H
    if (Rec.EpicenterToCurrPointF > 0) and (Rec.TravelTime > 0) and (Rec.SDistance > 0) then begin
      // <= 0 B   > 0 cb
      if Rec.EpicenterToCurrPointF <= Rec.SDistance then
        Rec.ArrivalSec := 0
      else
        Rec.ArrivalSec := Round(Rec.EpicenterToCurrPointF / (Rec.SDistance / Rec.TravelTime)) - Rec.TravelTime;
    end else
      Rec.ArrivalSec := 999999;

    result := true;
  except
    on e: exception do
      AddToLog('TPSWavesAccess.CalcPSWavesWithEpicenter', e.Message);
  end;
end;



end.
